from enum import StrEnum, auto


class CLIEnvironment(StrEnum):
    prod = auto()
    dev = auto()


def get_api_url(env: CLIEnvironment) -> str:
    url_lookup = {
        CLIEnvironment.prod: "https://api-ampere.jeffbrennan.dev",
        CLIEnvironment.dev: "http://127.0.0.1:8000",
    }
    return url_lookup[env]


def get_flag_emoji(country_code: str) -> str:
    """Converts a two-letter country code to a flag emoji."""
    code_points = [ord(c) + 127397 for c in country_code.upper()]
    return "".join(chr(i) for i in code_points)


def get_pct_change(last_period: int, this_period: int) -> float:
    if last_period == 0:
        return 100.0
    return (this_period - last_period) / last_period * 100



import datetime
import time
from typing import Optional

import pandas as pd
from dotenv import load_dotenv
from google.cloud import bigquery

from ampere.common import (
    DeltaTableWriteMode,
    DeltaWriteConfig,
    get_backend_db_con,
    get_current_time,
    get_model_primary_key,
    write_delta_table,
)
from ampere.models import PyPIDownload, PyPIQueryConfig


def record_pypi_query(query: PyPIQueryConfig) -> None:
    config = DeltaWriteConfig(
        table_dir="bronze",
        table_name=PyPIQueryConfig.__tablename__,  # pyright: ignore [reportArgumentType]
        pks=get_model_primary_key(PyPIQueryConfig),
        mode=DeltaTableWriteMode.APPEND,
    )

    write_delta_table([query], config)


def get_pypi_downloads_from_bigquery(
    config: PyPIQueryConfig, dry_run: bool = True
) -> Optional[pd.DataFrame]:
    max_date_where = ""
    if config.max_date is not None:
        max_date_where = (
            f"and TIMESTAMP_TRUNC(timestamp, day) <= timestamp ('{config.max_date}')"
        )

    cmd = f"""
        select
            project,
            timestamp_trunc(`timestamp`, hour)          as `timestamp`,
            coalesce(country_code, 'unknown')           as `country_code`,
            coalesce(file.version, 'unknown')           as `package_version`,
            coalesce(details.python, 'unknown')         as `python_version`,
            coalesce(details.distro.name, 'unknown')    as `system_distro_name`,
            coalesce(details.distro.version, 'unknown') as `system_distro_version`,
            coalesce(details.system.name, 'unknown')    as `system_name`,
            coalesce(details.system.release, 'unknown') as `system_release`,
            count(*)                                    as download_count,
            current_timestamp()                         as retrieved_at
        from `bigquery-public-data.pypi.file_downloads`
        where 
            TIMESTAMP_TRUNC(timestamp, day) >= timestamp ('{config.min_date}') 
            {max_date_where} 
            and project = '{config.repo}'
        group by all
        """

    print(cmd)
    if dry_run:
        return None

    client = bigquery.Client()
    query_job = client.query_and_wait(cmd)
    start_time = time.time()
    results = query_job.to_dataframe()

    elapsed_time = time.time() - start_time
    print(f"query finished in {elapsed_time:.2f} seconds")

    return results


def refresh_pypi_downloads_from_bigquery(
    query_config: PyPIQueryConfig, write_config: DeltaWriteConfig, dry_run: bool = True
) -> int:
    results = get_pypi_downloads_from_bigquery(query_config, dry_run)
    if results is None:
        print("dry run - exiting early")
        return 0

    if len(results) == 0:
        print(
            f"0 downloads found for time period: {query_config.min_date} - {query_config.max_date}"
        )
        record_pypi_query(query_config)
        return 0
    write_delta_table(results, write_config)

    record_pypi_query(query_config)
    return len(results)


def get_pypi_download_query_dates() -> list[PyPIQueryConfig]:
    max_query_days = 45

    con = get_backend_db_con()

    query_dates = con.sql(
        """
        select repo, max(max_date) as data_copied_through_date
        from pypi_download_queries
        group by repo
        """
    ).fetchall()

    # inconsistent bigquery data usage when querying current day data
    yesterday = datetime.datetime.now() - datetime.timedelta(days=1)
    max_date = datetime.datetime.strftime(yesterday, "%Y-%m-%d")

    queries = []
    for record in query_dates:
        repo = record[0]
        min_date_str = record[1]
        min_date = datetime.datetime.strptime(min_date_str, "%Y-%m-%d")
        days_to_query = yesterday - min_date

        if days_to_query.days > max_query_days:
            raise ValueError(
                f"check `pypi_download_queries` table for {record[0]} - newest date: {min_date_str}"
            )

        if days_to_query.days < 1:
            print(f"{repo} has data through {min_date_str}. skipping")
            continue

        queries.append(
            PyPIQueryConfig(
                repo=repo,
                min_date=min_date_str,
                max_date=max_date,
                retrieved_at=get_current_time(),
            )
        )
    return queries


def get_repos_with_releases() -> list[str]:
    con = get_backend_db_con()
    records = con.sql(
        """
        with
            release_repos as (
                select distinct
                    repo_id
                from stg_releases
            ),
            repo_details as (
                select
                    a.repo_id,
                    a.repo_name,
                    unnest(a.language) as "language"
                from stg_repos               a
                inner join release_repos b
                on a.repo_id = b.repo_id
            )
        select
            repo_name
        from repo_details
        where
            language.name = 'Python'   
        """
    ).fetchall()
    print(records)

    return [i[0] for i in records]


def get_backfill_queries(
    repo: str,
    min_date: datetime.datetime,
    max_date: Optional[datetime.datetime] = None,
    max_days_per_chunk: int = 15,
) -> list[PyPIQueryConfig]:
    max_date_final = max_date

    if max_date_final is None:
        n_days_to_fill = (get_current_time() - min_date).days
    else:
        n_days_to_fill = (max_date_final - min_date).days

    queries = []
    if n_days_to_fill < max_days_per_chunk:
        max_date = min_date + datetime.timedelta(days=n_days_to_fill)
        queries.append(
            PyPIQueryConfig(
                repo=repo,
                min_date=datetime.datetime.strftime(min_date, "%Y-%m-%d"),
                max_date=datetime.datetime.strftime(max_date, "%Y-%m-%d"),
                retrieved_at=get_current_time(),
            )
        )
        return queries

    chunks = n_days_to_fill // max_days_per_chunk + 1
    for _ in range(1, chunks):
        max_date = min_date + datetime.timedelta(days=max_days_per_chunk)
        queries.append(
            PyPIQueryConfig(
                repo=repo,
                min_date=datetime.datetime.strftime(min_date, "%Y-%m-%d"),
                max_date=datetime.datetime.strftime(max_date, "%Y-%m-%d"),
                retrieved_at=get_current_time(),
            )
        )
        min_date = max_date

    if max_date_final is not None:
        return queries

    yesterday = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=1)
    if max_date is not None and max_date > yesterday:
        return queries

    max_date = datetime.datetime.strftime(yesterday, "%Y-%m-%d")  # type: ignore
    queries.append(
        PyPIQueryConfig(
            repo=repo,
            min_date=datetime.datetime.strftime(min_date, "%Y-%m-%d"),
            max_date=max_date,  # type: ignore
            retrieved_at=get_current_time(),
        )
    )
    return queries


def add_backfill_to_table(
    repo: str,
    min_date: datetime.datetime,
    max_date: Optional[datetime.datetime],
    max_days_per_chunk: int,
    dry_run: bool,
):
    queries = get_backfill_queries(repo, min_date, max_date, max_days_per_chunk)

    print(f"backfilling {repo} {'-' * 20}")
    for query in queries:
        print(query.min_date, "->", query.max_date)

    refresh_all_pypi_downloads(queries, dry_run)


def refresh_all_pypi_downloads(
    queries: Optional[list[PyPIQueryConfig]] = None, dry_run: bool = True
) -> int:
    load_dotenv()

    write_config = DeltaWriteConfig(
        table_dir="bronze",
        table_name=PyPIDownload.__tablename__,  # pyright: ignore [reportArgumentType]
        pks=get_model_primary_key(PyPIDownload),
        mode=DeltaTableWriteMode.APPEND,  # less resource intensive than merge
    )

    if queries is None:
        queries = get_pypi_download_query_dates()

    if len(queries) == 0:
        print("nothing to write! exiting early")
        return 0

    records_added = 0
    for query in queries:
        records_added += refresh_pypi_downloads_from_bigquery(
            query, write_config, dry_run
        )
    return records_added


if __name__ == "__main__":
    refresh_all_pypi_downloads(dry_run=True)



import datetime
from dataclasses import dataclass
from typing import Any

from fastapi import APIRouter, Query, Request
from fastapi.exceptions import RequestValidationError
from pydantic import TypeAdapter

from ampere.api.limiter import limiter
from ampere.cli.common import CLIEnvironment
from ampere.common import get_frontend_db_con
from ampere.models import (
    FeedBounds,
    FeedPublic,
    FeedPublicAction,
    FeedPublicEvent,
    FeedPublicRecord,
    create_repo_enum,
)

router = APIRouter(prefix="/feed", tags=["feed"])


@dataclass
class FeedGroups:
    repo: str | None = None
    event: str | None = None
    action: str | None = None
    username: str | None = None


def apply_group_filters(
    base_query: str,
    params: list[Any],
    groups: FeedGroups,
) -> tuple[str, list[Any]]:
    if groups.repo is not None:
        valid_repos = create_repo_enum(CLIEnvironment.dev, False)
        try:
            repo_validated = valid_repos(groups.repo)  # type: ignore
        except ValueError:
            raise ValueError(f"Invalid repo: {groups.repo}")

        base_query += " and repo_name = ?"
        params.append(repo_validated.value)

    if groups.event is not None:
        base_query += " and event_type = ?"
        params.append(groups.event)

    if groups.action is not None:
        base_query += " and event_action = ?"
        params.append(groups.action)

    if groups.username is not None:
        base_query += " and lower(user_name) = ?"
        params.append(groups.username)

    return base_query, params


@router.get("/bounds", response_model=FeedBounds)
@limiter.limit("60/minute")
def get_feed_bounds(
    request: Request,
    repo: str | None = None,
    event: FeedPublicEvent | None = None,
    action: FeedPublicAction | None = None,
    username: str | None = None,
):
    con = get_frontend_db_con()

    repo_name = "'overall'" if repo is None else "repo_name"
    event_type = "'overall'" if event is None else "event_type"
    event_action = "'overall'" if action is None else "event_action"
    user_name = "'overall'" if username is None else "user_name"

    base_query = f"""
    select
        {repo_name} as repo,
        {event_type} as event,
        {event_action} as action,
        {user_name} as username,
        min(event_timestamp) as min_date, 
        max(event_timestamp) as max_date 
    from mart_feed_events
    where 1 = 1
    """
    params = []
    base_query, params = apply_group_filters(
        base_query,
        params,
        FeedGroups(repo, event, action, username),
    )

    base_query += "group by all"
    print(base_query)
    result = con.sql(base_query, params=params).to_df().to_dict(orient="records")
    return FeedBounds.model_validate(result[0])


@router.get("/list", response_model=FeedPublic)
@limiter.limit("60/minute")
def read_feed(
    request: Request,
    repo: str | None = None,
    event: FeedPublicEvent | None = None,
    action: FeedPublicAction | None = None,
    username: str | None = None,
    n_days: int | None = Query(default=60, le=9999),
    min_date: str | None = Query(None, regex=r"\d{4}-\d{2}-\d{2}", format="date"),
    max_date: str | None = Query(None, regex=r"\d{4}-\d{2}-\d{2}", format="date"),
    limit: int = Query(default=50, le=10_000),
    descending: bool = Query(default=True),
) -> FeedPublic:
    con = get_frontend_db_con()
    sort_order = "desc" if descending else "asc"
    params = []

    if any([min_date, max_date]):
        n_days = None

    base_query = """
    select * from mart_feed_events
    where 1 = 1
    """

    bounds = get_feed_bounds(request, repo, event, action)
    if n_days is not None:
        requested_min_date = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(days=n_days)

        if requested_min_date < bounds.min_date:
            maximum_n_days = (
                datetime.datetime.now(datetime.timezone.utc) - bounds.min_date
            ).days
            err = ""
            err += f"requested min date {requested_min_date.strftime('%Y-%m-%d')} [{n_days} days] < "
            err += f"available min date {bounds.min_date.strftime('%Y-%m-%d')} | "
            err += f"please use a value <= {maximum_n_days} for `n_days`"
            raise RequestValidationError(err)
        base_query += f"and event_timestamp >= now() - interval {n_days} days"

    if min_date is not None:
        try:
            min_date_dt = datetime.datetime.strptime(min_date, "%Y-%m-%d").astimezone(
                datetime.timezone.utc
            )
        except ValueError:
            raise RequestValidationError("min_date must be in the format YYYY-MM-DD")

        if min_date_dt < bounds.min_date:
            raise RequestValidationError(f"min_date must be after {bounds.min_date}")
        print(min_date)
        base_query += f" and event_timestamp >= date '{min_date}'"

    if max_date is not None:
        try:
            max_date_dt = datetime.datetime.strptime(max_date, "%Y-%m-%d").astimezone(
                datetime.timezone.utc
            )
        except ValueError:
            raise RequestValidationError("max_date must be in the format YYYY-MM-DD")

        if max_date_dt > bounds.max_date:
            raise RequestValidationError(f"max_date must be before {bounds.max_date}")
        base_query += f" and event_timestamp <= date '{max_date}'"

    base_query, params = apply_group_filters(
        base_query,
        params,
        FeedGroups(repo, event, action, username),
    )

    base_query += f"""
    order by event_timestamp {sort_order}
    limit {limit}
    """

    print(base_query, params)

    result = con.sql(base_query, params=params).to_df().to_dict(orient="records")
    ta = TypeAdapter(list[FeedPublicRecord])
    return FeedPublic(data=ta.validate_python(result), count=len(result))



import copy

import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dash_table, html

from ampere.common import get_frontend_db_con, timeit
from ampere.styling import (
    ColumnInfo,
    ScreenWidth,
    get_ampere_dt_style,
    get_table_title_style,
    style_dt_background_colors_by_rank,
)


def create_issues_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
       select
           repo,
           author,
           title,
           body,
           "date",
           "days old",
           "comments"
       from mart_issues
       order by repo, "days old"
        """,
        ).to_df()
    return df


def create_issues_summary_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
        select repo,
            "open issues",
            "median issue age (days)",
            "new issues (this month)",
            "closed issues (this month)"
        from mart_issues_summary
        order by "open issues" desc
        """
        ).to_df()
    return df


def handle_col_widths(
    style_cell_conditional_incoming: list[dict], breakpoint_name: str
) -> list[dict]:
    style_cell_conditional = copy.deepcopy(style_cell_conditional_incoming)
    style_cell_conditional = [
        i for i in style_cell_conditional if "minWidth" not in str(i)
    ]

    col_widths = {
        "repo": 60,
        "author": 90,
        "date": 50,
        "days old": 35,
        "comments": 45,
        "title": 150,
        "body": 300,
    }

    small_col_widths = {
        "repo": 100,
        "author": 180,
        "date": 100,
        "days old": 80,
        "comments": 100,
        "title": 250,
        "body": 400,
    }

    xs_col_widths = {
        "repo": 100,
        "author": 100,
        "date": 105,
        "days old": 80,
        "comments": 100,
        "title": 250,
        "body": 400,
    }

    width_lookup = {
        "xl": col_widths,
        "lg": col_widths,
        "md": small_col_widths,
        "sm": small_col_widths,
        "xs": xs_col_widths,
    }

    width_adjustment = [
        {
            "if": {"column_id": i},
            "minWidth": width_lookup[breakpoint_name][i],
            "maxWidth": width_lookup[breakpoint_name][i],
        }
        for i in [
            "repo",
            "author",
            "date",
            "days old",
            "comments",
            "title",
            "body",
        ]
    ]

    style_cell_conditional.extend(width_adjustment)
    return style_cell_conditional


def handle_summary_col_widths(
    style_cell_conditional: list[dict], breakpoint_name: str
) -> list[dict]:
    style_cell_conditional = [
        i for i in style_cell_conditional if "minWidth" not in str(i)
    ]
    if breakpoint_name in [ScreenWidth.xl, ScreenWidth.lg]:
        col_width = 50
    elif breakpoint_name in [ScreenWidth.md, ScreenWidth.sm]:
        col_width = 60
    else:
        col_width = 90

    style_cell_conditional.append(
        {"if": {"column_id": "repo"}, "minWidth": col_width, "maxWidth": col_width}
    )
    return style_cell_conditional


def handle_table_margins(style_table_incoming: dict, breakpoint_name: str) -> dict:
    style_table = copy.deepcopy(style_table_incoming)
    if breakpoint_name != "xl":
        style_table.pop("maxWidth", None)
        style_table.pop("width", None)
        style_table.pop("marginLeft", None)
        return style_table

    margin_adjustments = {"maxWidth": "65vw", "width": "65vw", "marginLeft": "12vw"}
    style_table.update(margin_adjustments)
    return style_table


def handle_title_margins(style_incoming: dict, breakpoint_name: str) -> dict:
    margin_adjustments = {
        "xl": {"maxWidth": "65vw", "width": "65vw", "marginLeft": "12vw"},
        "other": {"maxWidth": "90vw", "width": "90vw", "marginLeft": "0vw"},
    }

    style = copy.deepcopy(style_incoming)
    if breakpoint_name == "xl":
        margin_adjustment = margin_adjustments["xl"]
    else:
        margin_adjustment = margin_adjustments["other"]

    style.update(margin_adjustment)

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        style["fontSize"] = "16px"

    return style


@callback(
    [
        Output("summary-title", "children"),
        Output("summary-title", "style"),
        Output("issues-fade", "is_in"),
    ],
    [
        Input("summary-table", "children"),
        Input("breakpoints", "widthBreakpoint"),
        Input("color-mode-switch", "value"),
    ],
)
def display_summary_title(_, breakpoint_name, dark_mode: bool):
    table_title_style = get_table_title_style(dark_mode)
    summary_title = html.Label(
        "summary", style=handle_title_margins(table_title_style, breakpoint_name)
    )

    return summary_title, {}, True


@callback(
    [
        Output("summary-table", "children"),
        Output("summary-table", "style"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
@timeit
def get_styled_issues_summary_table(dark_mode: bool, breakpoint_name: str):
    summary_df = create_issues_summary_table()
    summary_data = summary_df.to_dict("records")
    n_repos = summary_df.shape[0]

    summary_style = get_ampere_dt_style(dark_mode)
    del summary_style["style_table"]["maxHeight"]
    del summary_style["style_table"]["height"]

    formatting_cols = [
        ColumnInfo(name="open issues", ascending=True, palette="oranges"),
        ColumnInfo(name="median issue age (days)", ascending=True, palette="oranges"),
        ColumnInfo(name="new issues (this month)", ascending=True, palette="oranges"),
        ColumnInfo(name="closed issues (this month)", ascending=True, palette="greens"),
    ]


    col_value_heatmaps = style_dt_background_colors_by_rank(
        df=summary_df,
        n_bins=n_repos,
        cols=formatting_cols,
        dark_mode=dark_mode,
    )

    summary_style["style_data_conditional"] = [
        i for i in summary_style["style_data_conditional"] if "odd" not in str(i)
    ]
    summary_style["style_data_conditional"].extend(col_value_heatmaps)

    summary_style["style_cell_conditional"] = handle_summary_col_widths(
        summary_style["style_cell_conditional"], breakpoint_name
    )

    summary_style["style_table"] = handle_table_margins(
        summary_style["style_table"], breakpoint_name
    )

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        summary_style["style_table"]["font_size"] = "12px"

    tbl = (
        dash_table.DataTable(
            data=summary_data,
            columns=[
                {"id": x, "name": x, "presentation": "markdown"}
                if x in ["repo"]
                else {"id": x, "name": x}
                for x in summary_df.columns
            ],
            **summary_style,
        ),
    )
    return tbl, {}


@callback(
    [
        Output("issues-title", "children"),
        Output("issues-title", "style"),
    ],
    [
        Input("issues-table", "children"),
        Input("breakpoints", "widthBreakpoint"),
        Input("color-mode-switch", "value"),
    ],
)
def display_issues_title(_, breakpoint_name, dark_mode: bool):
    table_title_style = get_table_title_style(dark_mode)
    issues_title = html.Label(
        "issues",
        style=handle_title_margins(table_title_style, breakpoint_name),
    )

    return issues_title, {}


@callback(
    [
        Output("issues-table", "children"),
        Output("issues-table", "style"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
@timeit
def get_styled_issues_table(dark_mode: bool, breakpoint_name: str):
    issues_df = create_issues_table()
    base_style = get_ampere_dt_style(dark_mode)
    base_style["style_cell_conditional"] = handle_col_widths(
        base_style["style_cell_conditional"], breakpoint_name
    )

    base_style["style_table"] = handle_table_margins(
        base_style["style_table"], breakpoint_name
    )

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        base_style["style_table"]["font_size"] = "12px"

    tbl = (
        dash_table.DataTable(
            issues_df.to_dict("records"),
            columns=[
                {"id": x, "name": x, "presentation": "markdown"}
                if x in ["repo", "author", "title", "body"]
                else {"id": x, "name": x}
                for x in issues_df.columns
            ],
            **base_style,
        ),
    )

    return tbl, {}


def layout():
    return dbc.Fade(
        id="issues-fade",
        children=[
            html.Div(id="summary-title", style={"visibility": "hidden"}),
            html.Div(id="summary-table", style={"visibility": "hidden"}),
            html.Br(),
            html.Div(id="issues-title", style={"visibility": "hidden"}),
            html.Div(id="issues-table", style={"visibility": "hidden"}),
        ],
        style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
        is_in=False,
    )



import datetime
from dataclasses import dataclass
from enum import StrEnum, auto
from functools import lru_cache
from typing import Optional

import requests
from sqlmodel import Field, SQLModel

from ampere.cli.common import CLIEnvironment, get_api_url


class Stargazer(SQLModel):
    __tablename__ = "stargazers"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    user_id: int = Field(primary_key=True, foreign_key="user.user_id")
    starred_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


@dataclass
class Language:
    name: str
    size_bytes: int


class Repo(SQLModel):
    __tablename__ = "repos"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True)
    repo_name: str
    license: Optional[str] = None
    topics: list[str]
    language: Optional[list[Language]] = None
    repo_size: int
    forks_count: int
    stargazers_count: int
    open_issues_count: int
    pushed_at: datetime.datetime
    created_at: datetime.datetime
    updated_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


class Fork(SQLModel):
    __tablename__ = "forks"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    fork_id: int = Field(primary_key=True)
    owner_id: int = Field(primary_key=True, foreign_key="user.user_id")
    created_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


# https://docs.github.com/en/rest/commits/commits?apiVersion=2022-11-28#list-commits
# https://docs.github.com/en/rest/commits/commits?apiVersion=2022-11-28#get-a-commit
@dataclass
class CommitStats:
    filename: str
    additions: int
    deletions: int
    changes: int
    status: str


class Commit(SQLModel):
    __tablename__ = "commits"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    commit_id: str = Field(primary_key=True)
    author_id: Optional[int] = Field(foreign_key="user.user_id")
    comment_count: int
    message: str
    stats: list[CommitStats]
    committed_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


# https://docs.github.com/en/rest/issues/issues?apiVersion=2022-11-28#list-repository-issues
class Issue(SQLModel):
    __tablename__ = "issues"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    issue_id: int = Field(primary_key=True)
    issue_number: int
    issue_title: str
    issue_body: Optional[str] = None
    author_id: int = Field(foreign_key="user.user_id")
    state: str
    state_reason: Optional[str] = None
    comments_count: int
    created_at: datetime.datetime
    updated_at: Optional[datetime.datetime] = None
    closed_at: Optional[datetime.datetime] = None
    retrieved_at: datetime.datetime = Field(primary_key=True)


# https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#list-pull-requests
class PullRequest(SQLModel):
    __tablename__ = "pull_requests"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    pr_id: int = Field(primary_key=True)
    pr_number: int
    pr_title: str
    pr_state: str
    pr_body: Optional[str] = None
    author_id: int = Field(foreign_key="user.user_id")
    created_at: datetime.datetime
    updated_at: datetime.datetime
    closed_at: Optional[datetime.datetime] = None
    merged_at: Optional[datetime.datetime] = None
    retrieved_at: datetime.datetime = Field(primary_key=True)


# https://docs.github.com/en/rest/users/users?apiVersion=2022-11-28#get-a-user
class User(SQLModel):
    __tablename__ = "users"  # pyright: ignore [reportAssignmentType]
    user_id: int = Field(primary_key=True)
    user_name: str
    full_name: Optional[str] = None
    company: Optional[str] = None
    avatar_url: str
    repos_count: int
    followers_count: int
    following_count: int
    created_at: datetime.datetime
    updated_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


# https://docs.github.com/en/rest/releases/releases?apiVersion=2022-11-28#list-releases
class Release(SQLModel):
    __tablename__ = "releases"  # pyright: ignore [reportAssignmentType]
    repo_id: int = Field(primary_key=True, foreign_key="repo.repo_id")
    release_id: int = Field(primary_key=True)
    release_name: str
    tag_name: str
    release_body: Optional[str] = None
    created_at: datetime.datetime
    published_at: datetime.datetime
    retrieved_at: datetime.datetime = Field(primary_key=True)


class Follower(SQLModel):
    __tablename__ = "followers"  # pyright: ignore [reportAssignmentType]
    user_id: int = Field(primary_key=True, foreign_key="users.user_id")
    follower_id: int = Field(primary_key=True, foreign_key="users.user_id")
    retrieved_at: datetime.datetime = Field(primary_key=True)


class PyPIDownload(SQLModel):
    __tablename__ = "pypi_downloads"  # pyright: ignore [reportAssignmentType]
    project: str = Field(primary_key=True, foreign_key="repo.repo_name")
    timestamp: datetime.datetime = Field(primary_key=True)
    country_code: str = Field(primary_key=True)
    package_version: str = Field(primary_key=True)
    python_version: str = Field(primary_key=True)
    system_distro_name: str = Field(primary_key=True)
    system_distro_version: str = Field(primary_key=True)
    system_name: str = Field(primary_key=True)
    system_release: str = Field(primary_key=True)
    download_count: int
    retrieved_at: datetime.datetime


# used to track which repos have been queried to prevent repeated date range queries on repos with no downloads
class PyPIQueryConfig(SQLModel):
    __tablename__ = "pypi_download_queries"  # pyright: ignore [reportAssignmentType]
    repo: str = Field(primary_key=True, foreign_key="repo.repo_name")
    retrieved_at: datetime.datetime = Field(primary_key=True)
    min_date: str
    max_date: Optional[str]


# viz model dataclases
@dataclass(slots=True, frozen=True)
class StargazerNetworkRecord:
    user_name: str
    followers_count: int
    starred_at: datetime.datetime
    retrieved_at: datetime.datetime
    repo_name: str


@dataclass(slots=True, frozen=True)
class Followers:
    user_id: str
    follower_id: int


@dataclass(slots=True, frozen=True)
class FollowerDetails:
    user_id: int
    user_name: str
    followers_count: int
    following_count: int
    followers: Optional[list[str]]
    following: Optional[list[str]]
    internal_followers_count: int
    internal_following_count: int
    internal_followers_pct: float
    internal_following_pct: float


# cli/api models
class DownloadsPublicGroup(StrEnum):
    overall = auto()
    country_code = auto()
    package_version = auto()
    python_version = auto()
    system_distro_name = auto()
    system_distro_version = auto()
    system_name = auto()
    system_release = auto()


class DownloadsGranularity(StrEnum):
    hourly = auto()
    daily = auto()
    weekly = auto()
    monthly = auto()


class DownloadsSummaryGranularity(StrEnum):
    weekly = auto()
    monthly = auto()


class FeedPublicEvent(StrEnum):
    commit = auto()
    fork = auto()
    issue = auto()
    pull_request = "pull request"
    star = auto()
    overall = auto()


class FeedPublicAction(StrEnum):
    created = auto()
    updated = auto()
    closed = auto()
    merged = auto()
    overall = auto()


class DownloadPublic(SQLModel):
    repo: str
    download_timestamp: datetime.datetime
    group_name: str
    group_value: str
    download_count: int


class DownloadsPublic(SQLModel):
    data: list[DownloadPublic]
    count: int


class ReposWithDownloads(SQLModel):
    repos: list[str]
    count: int


class GetDownloadsPublicConfig(SQLModel):
    granularity: DownloadsGranularity | DownloadsSummaryGranularity
    repo: str
    group: DownloadsPublicGroup
    n_days: int
    limit: int
    descending: bool


class FeedPublicRecord(SQLModel):
    repo_name: str
    user_name: str
    full_name: str | None = None
    event_id: str
    event_type: FeedPublicEvent
    event_action: FeedPublicAction
    event_data: str | None = None
    event_timestamp: datetime.datetime
    event_link: str | None = None


class FeedPublic(SQLModel):
    data: list[FeedPublicRecord]
    count: int


class FeedBounds(SQLModel):
    repo: str
    event: FeedPublicEvent
    action: FeedPublicAction
    min_date: datetime.datetime
    max_date: datetime.datetime


class ReposPublic(SQLModel):
    repos: list[str]
    count: int


def get_repos_with_downloads_dev() -> list[str]:
    from ampere.common import get_frontend_db_con

    with get_frontend_db_con() as con:
        repos = (
            con.sql(
                """
                select distinct a.repo 
                from mart_downloads_summary a 
                left join stg_repos b on a.repo = b.repo_name
                order by b.stargazers_count desc, a.repo
                """
            )
            .to_df()
            .squeeze()
            .tolist()
        )

    return repos


def get_repos_with_downloads_prod() -> list[str]:
    url = get_api_url(CLIEnvironment.prod)
    response = requests.get(f"{url}/downloads/repos")
    assert response.status_code == 200, print(response.json())
    model = ReposWithDownloads.model_validate(response.json())
    return model.repos


def get_repos_with_downloads(env: str) -> list[str]:
    if env == "dev":
        return get_repos_with_downloads_dev()
    return get_repos_with_downloads_prod()


def get_repo_names_dev() -> list[str]:
    from ampere.common import get_frontend_db_con

    with get_frontend_db_con() as con:
        repo_names = con.sql(
            "select repo_name from stg_repos order by stargazers_count desc"
        ).fetchall()

    return [repo_name[0] for repo_name in repo_names]


def get_repo_names_prod() -> list[str]:
    url = get_api_url(CLIEnvironment.prod)
    response = requests.get(f"{url}/repos/list")
    assert response.status_code == 200, print(response.json())
    model = ReposWithDownloads.model_validate(response.json())
    return model.repos


def get_repo_names(env: str) -> list[str]:
    if env == "dev":
        return get_repo_names_dev()
    return get_repo_names_prod()


@lru_cache()
def create_repo_enum(env: CLIEnvironment, with_downloads: bool) -> StrEnum:
    if with_downloads:
        repos = get_repos_with_downloads(env)
    else:
        repos = get_repo_names(env)

    return StrEnum("RepoEnum", {repo: repo for repo in repos})



import os
from typing import Any

import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dcc, html
from plotly.graph_objects import Figure

from ampere.app_shared import cache, update_tooltip
from ampere.cli.common import CLIEnvironment
from ampere.common import timeit
from ampere.models import get_repos_with_downloads
from ampere.styling import ScreenWidth
from ampere.viz import (
    read_dataframe_pickle,
    read_plotly_fig_pickle,
    viz_downloads,
)


@timeit
def get_viz_downloads(
    df_data: list[dict],
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
    group: str,
    repo: str,
    breakpoint_name: str,
) -> tuple[Figure, dict]:
    env = os.environ.get("AMPERE_ENV")

    if date_range == date_bounds and env == "prod":
        mode = "dark" if dark_mode else "light"
        f_name = f"downloads_{repo}_{group}_{mode}_{breakpoint_name}"
        try:
            fig = read_plotly_fig_pickle(f_name)
            print(f"obtained {group} fig from cache")
            return fig, {}

        except Exception as e:
            print(e)
            pass

    fig = viz_downloads(
        pd.DataFrame(df_data),
        group_name=group,
        date_range=date_range,
        dark_mode=dark_mode,
        screen_width=ScreenWidth(breakpoint_name),
    )

    return fig, {}


@cache.memoize()
@timeit
def dash_get_repos_with_downloads():
    return get_repos_with_downloads(CLIEnvironment.dev)


@callback(
    [
        Output("downloads-overall", "figure"),
        Output("downloads-overall", "style"),
    ],
    [
        Input("downloads-df", "data"),
        Input("date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("downloads-date-bounds", "data"),
        Input("repo-selection", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def viz_downloads_overall(
    df_data: list[dict],
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
    repo_name: str,
    breakpoint_name: str,
):
    return get_viz_downloads(
        df_data=df_data,
        date_range=date_range,
        dark_mode=dark_mode,
        date_bounds=date_bounds,
        group="overall",
        repo=repo_name,
        breakpoint_name=breakpoint_name,
    )


@callback(
    [
        Output("downloads-package-version", "figure"),
        Output("downloads-package-version", "style"),
    ],
    [
        Input("downloads-df", "data"),
        Input("date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("downloads-date-bounds", "data"),
        Input("repo-selection", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def viz_downloads_by_package_version(
    df_data: list[dict],
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
    repo_name: str,
    breakpoint_name: str,
):
    return get_viz_downloads(
        df_data=df_data,
        date_range=date_range,
        dark_mode=dark_mode,
        date_bounds=date_bounds,
        group="package_version",
        repo=repo_name,
        breakpoint_name=breakpoint_name,
    )


@callback(
    [
        Output("downloads-python-version", "figure"),
        Output("downloads-python-version", "style"),
    ],
    [
        Input("downloads-df", "data"),
        Input("date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("downloads-date-bounds", "data"),
        Input("repo-selection", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def viz_downloads_by_python_version(
    df_data: list[dict],
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
    repo_name: str,
    breakpoint_name: str,
):
    return get_viz_downloads(
        df_data=df_data,
        date_range=date_range,
        dark_mode=dark_mode,
        date_bounds=date_bounds,
        group="python_version",
        repo=repo_name,
        breakpoint_name=breakpoint_name,
    )


@callback(
    Output("downloads-fade", "is_in"),
    [
        Input("downloads-overall", "figure"),
        Input("downloads-package-version", "figure"),
        Input("downloads-python-version", "figure"),
    ],
)
def update_downloads_graph_fade(fig1, fig2, fig3):
    return all([fig1, fig2, fig3])


@callback(
    Output("downloads-df", "data"),
    Input("repo-selection", "value"),
)
@cache.memoize()
@timeit
def get_downloads_records(repo_name: str) -> list[dict]:
    print(f"cache miss: computing get_downloads_records for {repo_name}")
    df = read_dataframe_pickle(f"downloads_df_{repo_name}")
    return df.to_dict("records")


@callback(
    Output("date-slider", "tooltip"),
    Input("breakpoints", "widthBreakpoint"),
)
@timeit
def toggle_slider_tooltip_visibility(breakpoint_name: str) -> dict[Any, Any]:
    return update_tooltip(breakpoint_name)


@callback(
    [
        Output("date-slider", "min"),
        Output("date-slider", "max"),
        Output("date-slider", "value"),
        Output("date-slider", "marks"),
        Output("downloads-date-bounds", "data"),
    ],
    [
        Input("downloads-df", "data"),
    ],
)
@timeit
def get_downloads_records_date_ranges(df_data: list[dict]):
    df = pd.DataFrame(df_data)
    df["download_date"] = pd.to_datetime(df["download_date"])
    min_timestamp = df["download_date"].min().timestamp()
    max_timestamp = df["download_date"].max().timestamp()

    date_slider_value = [min_timestamp, max_timestamp]
    date_slider_marks = {
        min_timestamp: {"style": {"fontSize": 0}},
        max_timestamp: {"style": {"fontSize": 0}},
    }
    return (
        min_timestamp,
        max_timestamp,
        date_slider_value,
        date_slider_marks,
        [min_timestamp, max_timestamp],
    )


@callback(
    [
        Output("dl-repo-filter-width", "width"),
        Output("dl-date-filter-width", "width"),
        Output("dl-filter-padding-width", "width"),
        Output("dl-filter-row", "style"),
    ],
    Input("breakpoints", "widthBreakpoint"),
)
def update_filter_for_mobile(breakpoint_name: str):
    filter_style = {"top": "60px"}
    if breakpoint_name == ScreenWidth.xs:
        filter_style.update({"paddingTop": "20px"})
        return 4, 7, 1, filter_style

    if breakpoint_name == ScreenWidth.sm:
        return 3, 5, 4, filter_style

    if breakpoint_name == ScreenWidth.md:
        return 2, 4, 6, filter_style

    return 1, 4, 7, filter_style


@callback(
    Output("repo-selection", "style"),
    Input("breakpoints", "widthBreakpoint"),
)
def update_dropdown_font_size(breakpoint_name: str):
    if breakpoint_name == ScreenWidth.xs:
        return {"fontSize": "12px"}

    if breakpoint_name == ScreenWidth.sm:
        return {"fontSize": "14px"}

    return {"fontSize": "20px"}


def layout():
    return [
        dcc.Store("downloads-df"),
        dcc.Store("downloads-date-bounds"),
        dbc.Fade(
            id="downloads-fade",
            children=[
                dbc.Row(
                    children=[
                        dbc.Col(
                            dcc.Dropdown(
                                dash_get_repos_with_downloads(),
                                placeholder="quinn",
                                value="quinn",
                                clearable=False,
                                searchable=False,
                                id="repo-selection",
                            ),
                            id="dl-repo-filter-width",
                            style={"marginLeft": "5%"},
                        ),
                        dbc.Col(
                            html.Div(
                                dcc.RangeSlider(
                                    id="date-slider",
                                    step=86400,  # daily
                                    allowCross=True,
                                ),
                                style={
                                    "whiteSpace": "nowrap",
                                    "paddingLeft": "5%",
                                    "marginTop": "6px",
                                },
                            ),
                            id="dl-date-filter-width",
                        ),
                        dbc.Col(id="dl-filter-padding-width"),
                    ],
                    id="dl-filter-row",
                ),
                html.Br(),
                dcc.Graph(
                    "downloads-overall",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
                dcc.Graph(
                    "downloads-package-version",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
                dcc.Graph(
                    "downloads-python-version",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
            ],
            style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
            is_in=False,
        ),
    ]



from enum import StrEnum, auto

import typer

from ampere.cli.common import CLIEnvironment
from ampere.models import create_repo_enum


class CLIOutputFormat(StrEnum):
    json = auto()
    table = auto()


def repo_option_callback(
    value: str | None, ctx: typer.Context, with_downloads: bool
) -> str | None:
    if value is None:
        return None
    env: CLIEnvironment = ctx.obj.get("env", CLIEnvironment.prod)
    RepoEnum = create_repo_enum(env, with_downloads)

    try:
        return RepoEnum(value.lower())  # type: ignore
    except ValueError:
        valid_repos = [repo.name for repo in RepoEnum]  # type: ignore
        valid_repos_str = ""
        for i, repo in enumerate(valid_repos, 1):
            valid_repos_str += "\n" + f"{i}. {repo}"

        title = "Valid Repos"
        if with_downloads:
            title += " with Downloads"

        raise typer.BadParameter(f"\n{'='*20} {title} {'='*20}{valid_repos_str}")


def repo_callback_without_downloads(value: str | None, ctx: typer.Context) -> str | None:
    return repo_option_callback(value, ctx, with_downloads=False)


def repo_callback_with_downloads(value: str | None, ctx: typer.Context) -> str | None:
    return repo_option_callback(value, ctx, with_downloads=True)



from __future__ import annotations

import datetime
import pickle
from pathlib import Path
from typing import Any, Optional

import networkx as nx
import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go
import pypalettes
from plotly.graph_objs import Figure

from ampere.common import get_frontend_db_con, timeit
from ampere.get_repo_metrics import read_repos
from ampere.models import (
    FollowerDetails,
    Followers,
    StargazerNetworkRecord,
)
from ampere.styling import AmperePalette, ScreenWidth, get_ampere_colors


@timeit
def generate_repo_palette() -> dict[str, str]:
    with get_frontend_db_con() as con:
        repos = sorted(
            read_repos(con),
            key=lambda x: x.stargazers_count,
            reverse=True,
        )

    n_colors = 10
    n_repos = len(repos)
    repeats = (n_repos // n_colors) + 1

    colors = list(
        pypalettes.load_cmap("Tableau_10", cmap_type="discrete", repeat=repeats).rgb  # type: ignore
    )[0:n_repos]

    output = {}
    for i, repo in enumerate(repos):
        rgb_string = ", ".join(str(x) for x in colors[i])
        output[repo.repo_name] = f"rgb({rgb_string})"

    return output


def format_plot_name_list(
    names: list[str] | float | None, max_names: int = 5
) -> Optional[str]:
    if names is None or isinstance(names, float) or isinstance(names, int):
        return None

    names_clean = names[0 : min(max_names, len(names))]

    names_clean_str = ", ".join(names_clean)
    if len(names) > max_names:
        names_clean_str += "..."

    return names_clean_str


def read_pickle(pkl_name: str) -> Any:
    out_dir = Path(__file__).parents[1] / "data" / "viz"
    out_path = out_dir / f"{pkl_name}.pkl"
    with out_path.open("rb") as f:
        obj = pickle.load(f)
    return obj


def read_json(pkl_name: str) -> Any:
    out_dir = Path(__file__).parents[1] / "data" / "viz"
    out_path = out_dir / f"{pkl_name}.json"
    with out_path.open("r") as f:
        data = f.read()
    return data


def read_network_graph_pickle(pkl_name: str) -> nx.Graph:
    return read_pickle(pkl_name)


@timeit
def read_plotly_fig_pickle(pkl_name: str) -> Figure:
    return read_pickle(pkl_name)


@timeit
def read_dataframe_pickle(pkl_name: str) -> pd.DataFrame:
    return read_pickle(pkl_name)


@timeit
def read_plotly_fig_json(f_name: str) -> Figure:
    fig_data = read_json(f_name)
    return go.Figure(plotly.io.from_json(fig_data))


def filter_df_by_date_range(
    df: pd.DataFrame, col_name: str, date_range: Optional[list[int]] = None
) -> pd.DataFrame:
    if date_range is None:
        return df

    filter_date_min = datetime.datetime.fromtimestamp(date_range[0])
    filter_date_max = datetime.datetime.fromtimestamp(date_range[1])

    return df.query(f"{col_name} >= '{filter_date_min}'").query(
        f"{col_name} <= '{filter_date_max}'"
    )


def style_area_fig(fig: Figure, dark_mode: bool, screen_width: ScreenWidth) -> Figure:
    bg_color, font_color = get_ampere_colors(dark_mode, contrast=False)

    if screen_width == ScreenWidth.xs:
        legend_font_size = 12
        annotation_font_size = 14
        tick_font_size = 10

        fig.update_layout(
            legend=dict(
                title=None,
                itemsizing="constant",
                orientation="h",
                yanchor="top",
                y=1.475,
                xanchor="center",
                x=0.5,
                font=dict(size=legend_font_size, color=font_color),
            ),
            dragmode=False,
        )
    else:
        legend_font_size = 16
        annotation_font_size = 20
        tick_font_size = 16
        fig.update_layout(
            legend=dict(
                title=None,
                itemsizing="constant",
                font=dict(size=legend_font_size, color=font_color),
            )
        )
    fig.update_layout(paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    fig.for_each_annotation(
        lambda a: a.update(
            text="<b>" + a.text.split("=")[-1].replace("_", " ") + "</b>",
            font_size=annotation_font_size,
            font_color=font_color,
            borderpad=5,
            y=1.002,
        )
    )

    fig.for_each_yaxis(
        lambda y: y.update(
            title="",
            showline=True,
            linewidth=2,
            linecolor=font_color,
            color=font_color,
            mirror=True,
            tickfont_size=tick_font_size,
        )
    )
    fig.for_each_xaxis(
        lambda x: x.update(
            title="",
            showline=True,
            linewidth=2,
            linecolor=font_color,
            color=font_color,
            mirror=True,
            showticklabels=True,
            tickfont_size=tick_font_size,
        )
    )
    fixed_axes = screen_width == ScreenWidth.xs
    fig.update_yaxes(
        matches=None, showticklabels=True, showgrid=False, fixedrange=fixed_axes
    )
    fig.update_xaxes(showgrid=False, fixedrange=fixed_axes)

    fig.update_layout(
        title={
            "y": 0.95,
            "x": 0.5,
            "xanchor": "center",
            "yanchor": "top",
        },
    )

    if screen_width != ScreenWidth.xs:
        fig.update_layout(
            margin=dict(l=0, r=200, t=50, b=0),
        )
    else:
        fig.update_layout(
            margin=dict(l=0, r=0, t=0, b=0),
        )

    return fig


def viz_summary(
    df: pd.DataFrame,
    metric_type: str,
    date_range: Optional[list[int]] = None,
    screen_width: ScreenWidth = ScreenWidth.lg,
    dark_mode: bool = False,
) -> Figure:
    df_filtered = df.query(f"metric_type == '{metric_type}'")

    df_filtered = filter_df_by_date_range(df_filtered, "metric_date", date_range)

    template = "plotly_dark" if dark_mode else "plotly_white"
    repo_palette = generate_repo_palette()
    fig = px.area(
        df_filtered,
        x="metric_date",
        y="metric_count",
        color="repo_name",
        template=template,
        hover_name="repo_name",
        color_discrete_map=repo_palette,
        height=500,
        category_orders={"repo_name": repo_palette.keys()},
        facet_col="metric_type",  # single var facet col for plot title
    )

    fig = style_area_fig(fig, dark_mode, screen_width)
    return fig


def get_summary_data() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
        select
            repo_name,
            metric_type,
            metric_date,
            metric_count,
        from main.mart_repo_summary
        order by metric_date
    """,
        ).to_df()

    return df


@timeit
def get_downloads_data(repo_name: str) -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            f"""
            select
            repo,
            download_date,
            group_name,
            group_value,
            download_count
            from mart_downloads_summary
            where repo = '{repo_name}'
            order by download_date, download_count
            """,
        ).to_df()

    return df


@timeit
def viz_downloads(
    df: pd.DataFrame,
    group_name: str,
    date_range: Optional[list[int]] = None,
    dark_mode: bool = False,
    screen_width: ScreenWidth = ScreenWidth.lg,
) -> Figure:
    df_filtered = df.query(f"group_name=='{group_name}'")
    df_filtered = filter_df_by_date_range(df_filtered, "download_date", date_range)

    max_date = df_filtered["download_date"].max()
    categories = (
        df_filtered[(df_filtered["download_date"] == max_date)]
        .sort_values("download_count", ascending=False)["group_value"]
        .tolist()
    )

    template = "plotly_dark" if dark_mode else "plotly_white"
    fig = px.area(
        df_filtered,
        x="download_date",
        y="download_count",
        color="group_value",
        facet_col="group_name",
        color_discrete_sequence=px.colors.qualitative.T10,
        template=template,
        category_orders={"group_value": categories},
    )

    fig = style_area_fig(fig, dark_mode, screen_width)

    if group_name == "overall":
        fig.update_layout(showlegend=False)

    return fig


@timeit
def create_star_network_plot(
    graph: nx.Graph,
    repos: list[str],
    stargazers: list[StargazerNetworkRecord],
    dark_mode: bool,
    screen_width: ScreenWidth,
) -> go.Figure:
    if dark_mode:
        edge_color = "rgba(242, 240, 227, 0.3)"
        legend_text_color = AmperePalette.BRAND_TEXT_COLOR_DARK
    else:
        edge_color = "rgba(33, 33, 33, 0.3)"
        legend_text_color = AmperePalette.BRAND_TEXT_COLOR_LIGHT

    edge_x = []
    edge_y = []
    for edge in graph.edges():
        if edge[0] in repos:
            continue

        x0, y0 = graph.nodes[edge[0]]["pos"]
        x1, y1 = graph.nodes[edge[1]]["pos"]

        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])

    edge_trace = go.Scatter(
        x=edge_x,
        y=edge_y,
        line=dict(width=1, color=edge_color),
        hoverinfo="none",
        mode="lines",
        showlegend=False,
    )

    node_info = []
    for node in graph.nodes():
        node_data = graph.nodes.data()[node]
        repo = node_data["repo"]
        if node == repo:
            continue

        x, y = graph.nodes[node]["pos"]
        followers_count = node_data["followers_count"]
        user_name = node.split("_")[0]

        all_repos = [i.repo_name for i in stargazers if i.user_name == user_name]

        all_repos_text = ", ".join(all_repos)

        node_text_list = [
            f"<b> {user_name} </b>",
            "",
            f"followers: {followers_count}",
            f"repos: {all_repos_text}",
        ]

        node_text = "<br>".join(node_text_list)
        node_info.append(
            {
                "x": x,
                "y": y,
                "repo": repo,
                "text": node_text,
                "size": followers_count,
            }
        )

    node_df = pd.DataFrame(node_info)
    node_df["size_group"] = pd.qcut(node_df["size"], 6, labels=False, duplicates="drop")
    node_df["size_group"] = (node_df["size_group"] + 1) * 5

    repo_palette = generate_repo_palette()
    all_node_traces = []
    for repo in repos:
        repo_df = node_df.query(f"repo == '{repo}'")
        node_trace = go.Scatter(
            x=repo_df.x,
            y=repo_df.y,
            marker_size=repo_df.size_group,
            marker_color=repo_palette[repo],
            mode="markers",
            hoverinfo="text",
            hovertext=repo_df.text,
            name=repo,
        )
        all_node_traces.append(node_trace)

    fig = go.Figure(data=[edge_trace, *all_node_traces], layout=NETWORK_LAYOUT)

    legend_y = 1.3 if screen_width == ScreenWidth.xs else 1.02
    fig.update_layout(
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
        legend=dict(
            font=dict(color=legend_text_color, size=16),
            y=legend_y,
        ),
    )

    if screen_width == ScreenWidth.xs:
        fig.update_xaxes(fixedrange=True)
        fig.update_yaxes(fixedrange=True)
        fig.update_layout(dragmode=False)
    return fig


@timeit
def viz_star_network(dark_mode: bool, screen_width: ScreenWidth) -> Figure:
    with get_frontend_db_con() as con:
        stargazers = con.sql(
            """
        select
            user_name,
            followers_count,
            starred_at,
            retrieved_at,
            repo_name
        from int_network_stargazers,
    """,
        ).to_df()

    stargazers = list(StargazerNetworkRecord(*record) for record in stargazers.values)
    repos_with_stargazers = list(set(i.repo_name for i in stargazers))
    repo_palette = generate_repo_palette()
    repos = [i for i in repo_palette if i in repos_with_stargazers]

    network = read_network_graph_pickle("star_network")
    fig = create_star_network_plot(network, repos, stargazers, dark_mode, screen_width)
    return fig


@timeit
def create_follower_network_plot(
    graph: nx.Graph,
    follower_info: list[Followers],
    follower_details: dict[int, FollowerDetails],
    dark_mode: bool,
    screen_width: ScreenWidth,
) -> go.Figure:
    if dark_mode:
        edge_color = "rgba(242, 240, 227, 0.5)"
        legend_text_color = AmperePalette.BRAND_TEXT_COLOR_DARK
    else:
        edge_color = "rgba(33, 33, 33, 0.5)"
        legend_text_color = AmperePalette.BRAND_TEXT_COLOR_LIGHT

    all_connections = [(i.user_id, i.follower_id) for i in follower_info]
    solo_edges = {"x": [], "y": []}
    mutual_edges = {"x": [], "y": []}
    for edge in graph.edges():
        followed = edge[0]
        follower = edge[1]

        is_mutual = (follower, followed) in all_connections
        x0, y0 = graph.nodes[followed]["pos"]
        x1, y1 = graph.nodes[follower]["pos"]

        if is_mutual:
            mutual_edges["x"].extend([x0, x1, None])
            mutual_edges["y"].extend([y0, y1, None])
        else:
            solo_edges["x"].extend([x0, x1, None])
            solo_edges["y"].extend([y0, y1, None])

    solo_edge_trace = go.Scatter(
        x=solo_edges["x"],
        y=solo_edges["y"],
        line=dict(width=1, color=edge_color),
        hoverinfo="none",
        mode="lines",
        name="solo connection",
    )
    mutual_edge_trace = go.Scatter(
        x=mutual_edges["x"],
        y=mutual_edges["y"],
        line=dict(width=1, color="rgb(247, 111, 83)"),
        hoverinfo="none",
        mode="lines",
        name="mutual connection",
    )

    node_info = []
    for node in graph.nodes():
        x, y = graph.nodes[node]["pos"]
        node_details = follower_details[node]
        node_text_list = [
            "<b>" + node_details.user_name + "</b>",
            "",
            f"org followers: {node_details.internal_followers_count}/{node_details.followers_count} ({node_details.internal_followers_pct * 100:.02f}%)",
            f"org following: {node_details.internal_following_count}/{node_details.following_count} ({node_details.internal_following_pct * 100:.02f}%)",
        ]

        internal_followers_clean = format_plot_name_list(node_details.followers)
        internal_following_clean = format_plot_name_list(node_details.following)

        if internal_followers_clean is not None or internal_following_clean is not None:
            node_text_list += [""]

        if internal_followers_clean is not None:
            node_text_list += [f"followers: {internal_followers_clean}"]

        if internal_following_clean is not None:
            node_text_list += [f"following: {internal_following_clean}"]

        node_text = "<br>".join(node_text_list)
        node_info.append(
            {
                "x": x,
                "y": y,
                "text": node_text,
                "followers_count": node_details.followers_count,
                "internal_followers_count": node_details.internal_followers_count,
                "internal_followers_pct": node_details.internal_followers_pct,
                "following_count": node_details.following_count,
                "internal_following_count": node_details.internal_following_count,
                "internal_following_pct": node_details.internal_following_pct,
            }
        )

    node_df = pd.DataFrame(node_info)
    node_df["followers_group"] = pd.qcut(
        node_df["followers_count"],
        10,
        labels=False,
        duplicates="drop",
    )

    node_df["followers_group"] **= 10
    if dark_mode:
        node_df["followers_group"] = 10 - node_df["followers_group"]

    node_trace = go.Scatter(
        x=node_df.x,
        y=node_df.y,
        marker=dict(
            size=8,
            color=node_df.followers_group,
            colorscale="Greys",
            line=dict(width=1, color=legend_text_color),
        ),
        mode="markers",
        hoverinfo="text",
        hovertext=node_df.text,
        name="follower count",
    )

    fig = go.Figure(
        data=[solo_edge_trace, mutual_edge_trace, node_trace], layout=NETWORK_LAYOUT
    )

    fig.update_layout(
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
        legend=dict(font=dict(color=legend_text_color, size=16)),
    )

    if screen_width == ScreenWidth.xs:
        fig.update_xaxes(fixedrange=True)
        fig.update_yaxes(fixedrange=True)
        fig.update_layout(dragmode=False)
    return fig


@timeit
def viz_follower_network(dark_mode: bool, screen_width: ScreenWidth) -> Figure:
    with get_frontend_db_con() as con:
        followers = (
            con.sql("select user_id, follower_id from int_internal_followers")
            .to_df()
            .to_dict(orient="records")
        )

        follower_details_dict = (
            con.sql(
                """
        select
        user_id,
        user_name,
        followers_count,
        following_count,
        followers,
        following,
        internal_followers_count,
        internal_following_count,
        internal_followers_pct,
        internal_following_pct
        from int_network_follower_details
        """,
            )
            .to_df()
            .to_dict(orient="records")
        )

    followers = [Followers(**record) for record in followers]  # type: ignore
    follower_details_dc = [FollowerDetails(**i) for i in follower_details_dict]  # type: ignore
    follower_details = {i.user_id: i for i in follower_details_dc}

    # reads precomputed network graph from scheduled job
    network = read_network_graph_pickle("follower_network")

    fig = create_follower_network_plot(
        network, followers, follower_details, dark_mode, screen_width
    )
    return fig


NETWORK_LAYOUT = go.Layout(
    showlegend=True,
    hovermode="closest",
    xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
    yaxis=dict(
        showgrid=False,
        zeroline=False,
        showticklabels=False,
    ),
    template="none",
    margin=dict(t=0, l=0, r=0, b=0),
    legend=dict(
        title=None,
        itemsizing="constant",
        font=dict(size=14),
        orientation="h",
        yanchor="top",
        y=1.04,
        xanchor="center",
        x=0.5,
    ),
)



import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dash_table, html

from ampere.common import get_frontend_db_con
from ampere.styling import AmperePalette, ScreenWidth, get_ampere_dt_style


def create_feed_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """    
        select
            strftime(event_timestamp, '%Y-%m-%d')                                   as "date",
            strftime(event_timestamp, '%H:%M')                                      as "time",
            concat('[', user_name, ']', '(https://www.github.com/', user_name, ')') as "user",
            event_action                                                            as "action",
            event_type                                                              as "type",
            date_part('day', current_date - event_timestamp)                        as "days ago",
            repo_name                                                               as "repo",
            coalesce(event_data, '')                                                as "description",
            event_link
        from main.mart_feed_events
        order by "date" desc, "time" desc
        """,
        ).to_df()
    return df


def handle_table_margins(breakpoint_name: str):
    style_table = {}
    if breakpoint_name in ["lg", "xl"]:
        width_lookup = {"date": 80, "time": 45, "event": 400, "description": 400}
        style_table = {"maxWidth": "65vw", "width": "65vw", "marginLeft": "12vw"}

    elif breakpoint_name == "md":
        width_lookup = {"date": 100, "time": 60, "event": 400, "description": 400}

    elif breakpoint_name == "sm":
        width_lookup = {"date": 100, "time": 60, "event": 250, "description": 250}

    elif breakpoint_name == "xs":
        width_lookup = {"date": 100, "time": 60, "event": 200, "description": 200}
    else:
        raise ValueError(f"unhandled breakpoint name: {breakpoint_name}")

    margin_adjustment = [
        {
            "if": {"column_id": i},
            "minWidth": width_lookup[i],
            "maxWidth": width_lookup[i],
        }
        for i in ["date", "time", "event", "description"]
    ]

    return style_table, margin_adjustment


@callback(
    [
        Output("feed-table", "children"),
        Output("feed-table", "style"),
        Output("feed-fade", "is_in"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def style_feed_table(dark_mode: bool, breakpoint_name: str):
    raw_df = create_feed_table()
    df = format_feed_table(raw_df)
    feed_style = get_ampere_dt_style(dark_mode)

    if dark_mode:
        event_background_colors = {
            "pull request": "#263302",
            "issue": "#6b303a",
            "star": "#54401b",
            "commit": "#15524e",
            "fork": "#3b2133",
        }
        event_color_border = AmperePalette.BRAND_TEXT_COLOR_DARK
    else:
        event_background_colors = {
            "pull request": "#d9e6b5",
            "issue": "#edb4bd",
            "star": "#e8d3a9",
            "commit": "#b6dedc",
            "fork": "#e1ccdb",
        }
        event_color_border = AmperePalette.BRAND_TEXT_COLOR_LIGHT

    color_styles = [
        {
            "if": {"filter_query": f"{{event}} contains '{k}'", "column_id": "event"},
            "backgroundColor": v,
            "borderBottom": f"1px solid {event_color_border}",
        }
        for k, v in event_background_colors.items()
    ]
    feed_style["style_data_conditional"].extend(color_styles)

    adjusted_style_table, adjusted_style_cell_conditional = handle_table_margins(
        breakpoint_name
    )

    feed_style["style_table"].update(adjusted_style_table)
    feed_style["style_cell_conditional"].extend(adjusted_style_cell_conditional)
    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        feed_style["style_cell"]["font_size"] = "12px"

    tbl = dash_table.DataTable(
        df.to_dict("records"),
        columns=[
            (
                {"id": x, "name": x, "presentation": "markdown"}
                if x in ["event", "description"]
                else {"id": x, "name": x}
            )
            for x in df.columns
        ],
        **feed_style,
    )
    return tbl, {}, True


def format_feed_table(df: pd.DataFrame) -> pd.DataFrame:
    df["type_link"] = "[" + df["type"] + "]" + "(" + df["event_link"] + ")"
    df["repo_link"] = (
        "[" + df["repo"] + "]" + "(https://github.com/mrpowers-io/" + df["repo"] + ")"
    )
    df["type_link"] = df["type_link"].fillna("star")

    df["event"] = (
        df["user"]
        + " "
        + df["action"]
        + " a "
        + df["type_link"]
        + " in "
        + df["repo_link"]
        + " "
        + df["days ago"].astype(str)
        + " days ago"
    )
    df["event"] = df["event"].str.replace("created a star in", "starred")
    df["event"] = df["event"].str.replace("created a [fork", "[forked")
    df["event"] = df["event"].str.replace("a [issue", "an [issue")
    df["event"] = df["event"].str.replace(" 0 days ago", " today")
    df["event"] = df["event"].str.replace(" 1 days ago", " yesterday")
    df.loc[df["type"] == "fork", "event"] = df["event"].str.replace(" in ", " ")

    df_final = df[["date", "time", "event", "description"]]
    if not isinstance(df_final, pd.DataFrame):
        raise TypeError()

    return df_final


def layout():
    return dbc.Fade(
        id="feed-fade",
        children=[
            html.Br(),
            html.Div(id="feed-table", style={"visibility": "hidden"}),
        ],
        style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
        is_in=False,
    )



import argparse
import os

import dash
import dash_bootstrap_components as dbc
import dash_breakpoints
from dash import Input, Output, State, callback, dcc, html

from ampere.app_shared import cache
from ampere.pages import (
    about,
    downloads,
    feed,
    issues,
    network_followers,
    network_stargazers,
    status,
    summary,
)
from ampere.styling import AmperePalette, ScreenWidth, get_ampere_colors


@callback(
    [
        Output("navbar-brand", "style"),
        Output("downloads-link", "style"),
        Output("feed-link", "style"),
        Output("issues-link", "style"),
        Output("network-link", "toggle_style"),
        Output("status-link", "style"),
        Output("about-link", "style"),
    ],
    [
        Input("current-url", "pathname"),
        Input("color-mode-switch", "value"),
    ],
)
def update_downloads_link_color(pathname: str, dark_mode: bool):
    _, color = get_ampere_colors(dark_mode, contrast=False)
    highlighted_background_color, highlighted_text_color = get_ampere_colors(
        dark_mode, contrast=True
    )

    pages = ["", "downloads", "feed", "issues", "network", "status", "about"]
    output_styles = [{"color": color} for _ in range(len(pages))]

    if pathname in ["network"]:
        return output_styles

    current_page = pathname.removeprefix("/").split("-")[0]

    output_styles[pages.index(current_page)] = {
        "color": highlighted_text_color,
        "backgroundColor": highlighted_background_color,
        "borderRadius": "20px",
    }

    return output_styles


@callback(
    Output("navbar-collapse", "is_open"),
    [Input("navbar-toggler", "n_clicks")],
    [State("navbar-collapse", "is_open")],
)
def toggle_navbar_collapse(n_clicks, is_open):
    if n_clicks:
        return not is_open
    return is_open


@callback(
    Output("navbar-collapse", "is_open", allow_duplicate=True),
    Input("current-url", "pathname"),
    prevent_initial_call=True,
)
def close_navbar_on_navigate(_):
    return False


@callback(
    [
        Output("color-mode-switch", "children"),
        Output("color-mode-switch", "value"),
    ],
    Input("color-mode-switch", "n_clicks"),
    State("color-mode-switch", "children"),
)
def toggle_color_mode(n_clicks, _):
    is_dark = n_clicks % 2 == 1
    if is_dark:
        return html.I(
            className="fas fa-sun",
            style={"color": AmperePalette.PAGE_BACKGROUND_COLOR_LIGHT},
        ), True

    return html.I(
        className="fas fa-moon",
        style={"color": AmperePalette.PAGE_BACKGROUND_COLOR_DARK},
    ), False


@callback(
    [
        Output("ampere-page", "className"),
        Output("navbar", "className"),
    ],
    Input("color-mode-switch", "value"),
)
def toggle_page_color(dark_mode: bool):
    class_name = "dark-mode" if dark_mode else "light-mode"
    return class_name, class_name


def layout():
    navbar = dbc.Navbar(
        dbc.Container(
            [
                dbc.NavbarBrand(
                    "ampere", href="/", class_name="navbar-brand", id="navbar-brand"
                ),
                dbc.NavbarToggler(id="navbar-toggler"),
                dbc.Collapse(
                    dbc.Nav(
                        [
                            dcc.Location(id="current-url", refresh=False),
                            dbc.NavItem(
                                dbc.NavLink(
                                    id="downloads-link",
                                    children="downloads",
                                    href="downloads",
                                    class_name="downloads-link",
                                )
                            ),
                            dbc.NavItem(
                                dbc.NavLink(id="feed-link", children="feed", href="feed")
                            ),
                            dbc.NavItem(
                                dbc.NavLink(
                                    id="issues-link", children="issues", href="issues"
                                )
                            ),
                            dbc.DropdownMenu(
                                children=[
                                    dbc.DropdownMenuItem(
                                        id="stargazers-dropdown",
                                        children="stargazers",
                                        href="network-stargazers",
                                    ),
                                    dbc.DropdownMenuItem(
                                        id="followers-dropdown",
                                        children="followers",
                                        href="network-followers",
                                    ),
                                ],
                                id="network-link",
                                nav=True,
                                in_navbar=True,
                                label="networks",
                            ),
                            dbc.NavItem(
                                dbc.NavLink(
                                    id="status-link", children="status", href="status"
                                )
                            ),
                            dbc.NavItem(
                                dbc.NavLink(
                                    id="about-link", children="about", href="about"
                                )
                            ),
                        ],
                        className="ml-auto",
                        navbar=True,
                    ),
                    id="navbar-collapse",
                    navbar=True,
                ),
                dbc.NavItem(
                    dbc.Button(
                        id="color-mode-switch",
                        n_clicks=0,
                        children=html.I(
                            className="fas fa-moon",
                            style={
                                "color": AmperePalette.PAGE_BACKGROUND_COLOR_DARK,
                            },
                        ),
                        color="link",
                    )
                ),
            ],
            fluid=True,
            id="navbar-container",
        ),
        id="navbar",
    )

    return dbc.Container(
        id="ampere-page",
        children=[
            navbar,
            html.Br(),
            dash_breakpoints.WindowBreakpoints(
                id="breakpoints",
                widthBreakpointThresholdsPx=[
                    768,
                    1200,
                    1920,
                    2560,
                ],
                widthBreakpointNames=[i.value for i in ScreenWidth],
            ),
            dash.page_container,
        ],
        fluid=True,
    )


def init_app(env: str = "prod"):
    serve_locally = {"dev": True, "prod": False}[env]

    app = dash.Dash(
        use_pages=True,
        external_stylesheets=["assets/css/bootstrap.min.css", dbc.icons.FONT_AWESOME],
        suppress_callback_exceptions=True,
        compress=True,
        serve_locally=serve_locally,
    )

    app.index_string = f"""
    <!DOCTYPE html>
    <html>
        <head>
            {{%metas%}}
            <title>{{%title%}}</title>
            {{%favicon%}}
            {{%css%}}
            <style>
                body, #navbar {{
                    background-color: {AmperePalette.PAGE_BACKGROUND_COLOR_LIGHT} !important;
                    margin: 0;
                }}
                
            </style>
        </head>
        <body>
            {{%app_entry%}}
            <footer>
                {{%config%}}
                {{%scripts%}}
                {{%renderer%}}
            </footer>
        </body>
    </html>
    """
    server = app.server
    cache.init_app(server)

    app.layout = layout()

    dash.register_page(summary.__name__, name="summary", path="/", layout=summary.layout)
    dash.register_page(downloads.__name__, name="downloads", layout=downloads.layout)
    dash.register_page(feed.__name__, name="feed", layout=feed.layout)
    dash.register_page(issues.__name__, name="issues", layout=issues.layout)

    dash.register_page(network_followers.__name__, layout=network_followers.layout)
    dash.register_page(network_stargazers.__name__, layout=network_stargazers.layout)

    dash.register_page(status.__name__, name="status", layout=status.layout)
    dash.register_page(about.__name__, name="about", layout=about.layout)

    return app, server


def run_app(app):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env",
        type=str,
        default="prod",
        help="Environment to run the app in. Default is prod",
    )
    envs = {
        "prod": {"host": "0.0.0.0", "debug": False},
        "dev": {"host": "127.0.0.1", "debug": True},
    }

    env = parser.parse_args().env
    os.environ["AMPERE_ENV"] = env

    app.run(host=envs[env]["host"], debug=envs[env]["debug"])


app, server = init_app()

if __name__ == "__main__":
    run_app(app)



from dagster import Definitions
from dagster_dbt import DbtCliResource

from .assets import (
    ampere_dbt_assets,
    bigquery_table_copy,
    dagster_get_commits,
    dagster_get_followers,
    dagster_get_following,
    dagster_get_forks,
    dagster_get_issues,
    dagster_get_pull_requests,
    dagster_get_pypi_downloads,
    dagster_get_releases,
    dagster_get_repos,
    dagster_get_stargazers,
    dagster_get_users,
    dagster_refresh_downloads_plots,
    dagster_refresh_follower_network,
    dagster_refresh_star_network,
    dagster_refresh_summary_plots,
    dagster_test_run_fail,
    dagster_test_run_fail2,
    dagster_test_run_pass,
    github_metrics_table_copy,
)
from .jobs import bigquery_daily_job, github_metrics_daily_4_job
from .project import ampere_project
from .schedules import schedules
from .sensors import email_on_run_failure

defs = Definitions(
    assets=[
        ampere_dbt_assets,
        dagster_get_stargazers,
        dagster_get_repos,
        dagster_get_releases,
        dagster_get_forks,
        dagster_get_issues,
        dagster_get_pull_requests,
        dagster_get_users,
        dagster_get_commits,
        dagster_get_followers,
        dagster_get_following,
        dagster_get_pypi_downloads,
        dagster_refresh_star_network,
        dagster_refresh_follower_network,
        dagster_refresh_summary_plots,
        dagster_refresh_downloads_plots,
        github_metrics_table_copy,
        bigquery_table_copy,
        dagster_test_run_fail,
        dagster_test_run_fail2,
        dagster_test_run_pass,
    ],
    jobs=[github_metrics_daily_4_job, bigquery_daily_job],
    schedules=schedules,
    resources={
        "dbt": DbtCliResource(project_dir=ampere_project),
    },
    sensors=[email_on_run_failure],
)



import copy
import datetime
from dataclasses import dataclass
from enum import StrEnum, auto
from typing import Annotated

import requests
import typer
from pydantic import BaseModel
from rich import box
from rich.console import Console
from rich.table import Table

from ampere.cli.common import get_api_url
from ampere.cli.models import CLIOutputFormat, repo_callback_without_downloads
from ampere.models import FeedPublic, FeedPublicAction, FeedPublicEvent

console = Console()
feed_app = typer.Typer()


@dataclass
class ParsedDates:
    date_format: str
    dates: list[datetime.datetime]
    trunc_dates: list[datetime.datetime]
    formatted_dates: list[str]


class FeedSummary(BaseModel):
    date: str
    stars: str
    forks: str
    issues: str
    prs: str
    commits: str
    total: str


class FeedGranularity(StrEnum):
    daily = auto()
    weekly = auto()
    monthly = auto()


class FeedSummaryOutput(BaseModel):
    records: list[FeedSummary]
    min_date: datetime.datetime
    max_date: datetime.datetime
    grand_total: FeedSummary
    granularity: FeedGranularity
    repo: str | None
    event: FeedPublicEvent | None
    action: FeedPublicAction | None
    username: str | None
    n_periods: int | None


class FeedSummaryConfig(BaseModel):
    granularity: FeedGranularity
    repo: str | None
    event: FeedPublicEvent | None
    action: FeedPublicAction | None
    username: str | None
    n_periods: int | None
    descending: bool


def date_trunc(dt: datetime.datetime, granularity: FeedGranularity) -> datetime.datetime:
    if granularity == granularity.daily:
        return dt.replace(hour=0, minute=0, second=0, microsecond=0)
    elif granularity == granularity.weekly:
        start_of_week = dt - datetime.timedelta(days=dt.weekday())
        return start_of_week.replace(hour=0, minute=0, second=0, microsecond=0)
    elif granularity == granularity.monthly:
        return dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def get_feed_list_response(
    repo: str | None,
    event: FeedPublicEvent | None,
    action: FeedPublicAction | None,
    username: str | None,
    n_days: int | None,
    min_date: str | None,
    max_date: str | None,
    limit: int,
    descending: bool,
    ctx: typer.Context,
) -> FeedPublic:
    env = ctx.obj["env"]
    base_url = get_api_url(env)

    url = f"{base_url}/feed/list"
    username = username.lower() if username is not None else None
    params = {
        "repo": repo,
        "event": event,
        "action": action,
        "username": username,
        "n_days": n_days,
        "min_date": min_date,
        "max_date": max_date,
        "limit": limit,
        "descending": descending,
    }
    response = requests.get(url, params=params)
    assert response.status_code == 200, response.json()
    model = FeedPublic.model_validate(response.json())
    return model


def format_feed_output(model: FeedPublic) -> Table:
    title = "Feed Events"
    table = Table(title=title, title_justify="left", title_style="bold", box=box.ROUNDED)

    cols = ["timestamp", "repo", "user", "event", "action", "data", "link"]
    for col in cols:
        table.add_column(col)

    max_data_len = 25
    for item in model.data:
        if item.event_data is None:
            event_data = ""
        elif len(item.event_data) > max_data_len:
            event_data = item.event_data[0:max_data_len] + "..."
        else:
            event_data = item.event_data

        table.add_row(
            item.event_timestamp.strftime("%Y-%m-%dT%H:%M:%S"),
            item.repo_name,
            item.user_name,
            item.event_type,
            item.event_action,
            event_data,
            item.event_link,
        )
        table.add_section()

    return table


def format_feed_summary(model: FeedSummaryOutput) -> Table:
    title = "Feed Summary"
    subtitle = (
        f"{model.min_date.strftime('%Y-%m-%d')} to {model.max_date.strftime('%Y-%m-%d')}"
    )
    title_full = f"{title}\n{subtitle}"
    table = Table(
        title=title_full,
        title_justify="left",
        title_style="bold",
        box=box.ROUNDED,
    )

    date_col_lookup = {
        FeedGranularity.daily: "date",
        FeedGranularity.weekly: "week",
        FeedGranularity.monthly: "month",
    }

    cols = [
        date_col_lookup[model.granularity],
        "stars",
        "forks",
        "issues",
        "prs",
        "commits",
    ]

    for col in cols:
        table.add_column(col)

    table.add_column("total", justify="right")

    prev_date_divider = model.records[0].date.split("-")[-2]
    for i, record in enumerate(model.records):
        date_divider = record.date.split("-")[-2]
        if prev_date_divider != date_divider:
            table.add_section()
        prev_date_divider = date_divider

        table.add_row(
            record.date,
            record.stars,
            record.forks,
            record.issues,
            record.prs,
            record.commits,
            record.total,
        )

    table.add_section()
    table.add_row(
        model.grand_total.date,
        model.grand_total.stars,
        model.grand_total.forks,
        model.grand_total.issues,
        model.grand_total.prs,
        model.grand_total.commits,
        model.grand_total.total,
    )

    return table


def parse_summary_dates(
    model: FeedPublic,
    granularity: FeedGranularity,
    descending: bool,
    n_periods: int | None,
) -> ParsedDates:
    dates = sorted(set(i.event_timestamp for i in model.data))
    max_date = max(dates)
    min_date = min(dates)
    n_days_lookup = {
        FeedGranularity.daily: 1,
        FeedGranularity.weekly: 7,
        FeedGranularity.monthly: 30,
    }
    date_format = "%Y-%m" if granularity == FeedGranularity.monthly else "%Y-%m-%d"

    if n_periods is None:
        min_date = date_trunc(min_date, granularity)
        max_date = date_trunc(max_date, granularity)
        trunc_dates = [
            max_date - datetime.timedelta(days=n_days_lookup[granularity] * i)
            for i in range(1, 1 + (max_date - min_date).days)
        ]

        return ParsedDates(
            date_format=date_format,
            dates=dates,
            trunc_dates=trunc_dates,
            formatted_dates=[i.strftime(date_format) for i in trunc_dates],
        )

    trunc_dates = []
    for i in range(n_periods):
        expected_date = max_date - datetime.timedelta(days=n_days_lookup[granularity] * i)
        trunc_dates.append(date_trunc(expected_date, granularity))

    min_trunc_date = min(trunc_dates)
    trunc_dates = sorted(trunc_dates, reverse=descending)
    formatted_dates = [i.strftime(date_format) for i in trunc_dates]
    dates = [i for i in dates if i >= min_trunc_date]
    if min_trunc_date not in dates:
        dates.insert(0, min_trunc_date)

    return ParsedDates(
        date_format=date_format,
        dates=dates,
        trunc_dates=trunc_dates,
        formatted_dates=formatted_dates,
    )


def create_feed_summary(
    model: FeedPublic,
    config: FeedSummaryConfig,
) -> FeedSummaryOutput:
    counts = {
        "star": 0,
        "fork": 0,
        "issue": {
            "created": 0,
            "updated": 0,
            "closed": 0,
        },
        "pull request": {
            "created": 0,
            "updated": 0,
            "merged": 0,
            "closed": 0,
        },
        "commit": 0,
        "total": 0,
    }

    parsed_dates = parse_summary_dates(
        model,
        config.granularity,
        config.descending,
        config.n_periods,
    )

    counts_by_date = {
        date: copy.deepcopy(counts) for date in parsed_dates.formatted_dates
    }
    counts_by_date["grand_total"] = counts.copy()
    for record in model.data:
        record_date = date_trunc(record.event_timestamp, config.granularity).strftime(
            parsed_dates.date_format
        )
        if record_date not in counts_by_date:
            continue

        if record.event_type in ["star", "fork", "commit"]:
            counts_by_date[record_date][record.event_type] += 1
            counts_by_date[record_date]["total"] += 1

            counts_by_date["grand_total"][record.event_type] += 1
            counts_by_date["grand_total"]["total"] += 1
            continue

        counts_by_date[record_date][record.event_type][record.event_action] += 1
        counts_by_date[record_date]["total"] += 1

        counts_by_date["grand_total"][record.event_type][record.event_action] += 1
        counts_by_date["grand_total"]["total"] += 1

    summary_records = []
    grand_total_records = []
    for k, v in counts_by_date.items():
        pr = v["pull request"]
        prs = f"+{pr['created']} ~{pr['updated'] + pr['merged']} -{pr['closed']}"

        issues = (
            f"+{v['issue']['created']} ~{v['issue']['updated']} -{v['issue']['closed']}"
        )

        if k == "grand_total":
            date = "GRAND TOTAL"
        else:
            date = k

        summary = FeedSummary(
            date=date,
            stars=f'+{v["star"]}',
            forks=f'+{v["fork"]}',
            issues=issues,
            prs=prs,
            commits=f'+{v["commit"]}',
            total=f"{v['total']:,}",
        )

        if k == "grand_total":
            grand_total_records.append(summary)
        else:
            summary_records.append(summary)

    return FeedSummaryOutput(
        records=summary_records,
        min_date=min(parsed_dates.dates),
        max_date=max(parsed_dates.dates),
        grand_total=grand_total_records[0],
        repo=config.repo,
        event=config.event,
        action=config.action,
        username=config.username,
        n_periods=config.n_periods,
        granularity=config.granularity,
    )


@feed_app.command("list")
def list_feed(
    ctx: typer.Context,
    repo: Annotated[
        str | None, typer.Option("--repo", "-r", callback=repo_callback_without_downloads)
    ] = None,
    event: Annotated[FeedPublicEvent | None, typer.Option("--event", "-e")] = None,
    action: Annotated[FeedPublicAction | None, typer.Option("--action", "-a")] = None,
    username: Annotated[str | None, typer.Option("--user", "-u")] = None,
    n_days: Annotated[int | None, typer.Option("--n-days", "-n")] = None,
    min_date: Annotated[str | None, typer.Option("--min-date", "-mind")] = None,
    max_date: Annotated[str | None, typer.Option("--max-date", "-maxd")] = None,
    limit: Annotated[int, typer.Option("--limit", "-l")] = 50,
    descending: Annotated[bool, typer.Option("--desc/--asc", "-d/-a")] = True,
    output: Annotated[
        CLIOutputFormat, typer.Option("--output", "-o")
    ] = CLIOutputFormat.table,
) -> None:
    # remove default limit if min_date is set
    if min_date is not None and limit == 50:
        limit = 10_000

    model = get_feed_list_response(
        repo,
        event,
        action,
        username,
        n_days,
        min_date,
        max_date,
        limit,
        descending,
        ctx,
    )
    if len(model.data) == 0:
        console.print("No feed events found.")
        return

    if output == CLIOutputFormat.json:
        console.print_json(model.model_dump_json())
        return

    table = format_feed_output(model)
    console.print(table)


@feed_app.command("summary")
def summarize_feed(
    ctx: typer.Context,
    granularity: Annotated[
        FeedGranularity,
        typer.Option(
            "--granularity",
            "-g",
            prompt=True,
        ),
    ] = FeedGranularity.daily,
    repo: Annotated[
        str | None, typer.Option("--repo", "-r", callback=repo_callback_without_downloads)
    ] = None,
    event: Annotated[FeedPublicEvent | None, typer.Option("--event", "-e")] = None,
    action: Annotated[FeedPublicAction | None, typer.Option("--action", "-a")] = None,
    username: Annotated[str | None, typer.Option("--user", "-u")] = None,
    n_periods: Annotated[int | None, typer.Option("--n-periods", "-n")] = 30,
    min_date: Annotated[str | None, typer.Option("--min-date", "-mind")] = None,
    max_date: Annotated[str | None, typer.Option("--max-date", "-maxd")] = None,
    descending: Annotated[bool, typer.Option("--desc/--asc", "-d/-a")] = True,
    output: Annotated[
        CLIOutputFormat, typer.Option("--output", "-o")
    ] = CLIOutputFormat.table,
) -> None:
    day_lookup = {
        FeedGranularity.daily: 1,
        FeedGranularity.weekly: 7,
        FeedGranularity.monthly: 30,
    }
    if n_periods is None:
        n_days = None
    else:
        n_days = day_lookup[granularity] * n_periods

    model = get_feed_list_response(
        repo,
        event,
        action,
        username,
        n_days,
        min_date,
        max_date,
        10_000,
        descending,
        ctx,
    )
    if len(model.data) == 0:
        console.print("No feed events found.")
        return

    summary = create_feed_summary(model, FeedSummaryConfig(**locals()))

    if output == CLIOutputFormat.json:
        console.print_json(summary.model_dump_json())
        return

    table = format_feed_summary(summary)
    console.print(table)



from typing import Any

from flask_caching import Cache

from ampere.styling import ScreenWidth

# fixes circular import error when attempting to import cache from app.py
cache = Cache(
    config={
        "CACHE_TYPE": "simple",
        "CACHE_DEFAULT_TIMEOUT": 60,
    }
)


def update_tooltip(breakpoint_name: str) -> dict[Any, Any]:
    if breakpoint_name == ScreenWidth.xs:
        tooltip_font_size = "12px"
    else:
        tooltip_font_size = "16px"

    return {
        "placement": "bottom",
        "always_visible": True,
        "transform": "secondsToYMD",
        "style": {"fontSize": tooltip_font_size},
    }



from pathlib import Path

import polars as pl

from ampere.common import (
    DeltaTableWriteMode,
    DeltaWriteConfig,
    get_model_primary_key,
    write_delta_table,
)
from ampere.models import PyPIDownload


def main(dry_run: bool = True):
    table_name = str(PyPIDownload.__tablename__)
    write_config = DeltaWriteConfig(
        table_dir="bronze",
        table_name=table_name,
        pks=get_model_primary_key(PyPIDownload),
    )
    tbl_dir = Path(__file__).parents[1] / "data" / "bronze"
    tbl_path = tbl_dir / f"{table_name}CLONE"
    updated_df = pl.scan_delta(str(tbl_path)).with_columns(
        pl.col("retrieved_at")
        .cast(pl.Datetime)
        .dt.replace_time_zone("UTC")
        .alias("retrieved_at")
    )
    if dry_run:
        print(updated_df.collect().head())
        return
    chunk_size = 5_000_000
    offset = 0
    while True:
        chunked_df = updated_df.slice(offset, chunk_size).collect()
        n_records = chunked_df.shape[0]
        if n_records == 0:
            break

        if offset == 0:
            mode = DeltaTableWriteMode.OVERWRITE_WITH_SCHEMA
        else:
            mode = DeltaTableWriteMode.APPEND

        write_delta_table(
            chunked_df,
            write_config.table_dir,
            write_config.table_name,
            write_config.pks,
            cleanup=False,
            mode=mode,
        )
        offset += chunk_size

    print("done")


if __name__ == "__main__":
    main(dry_run=False)



from pathlib import Path

import duckdb

from ampere.common import get_backend_db_con, get_frontend_db_con, timeit


def create_new_frontend_db():
    base_dir = Path(__file__).parents[1] / "data"
    db_path = base_dir / "frontend.duckdb"

    print("deleting existing file...")
    db_path.unlink(missing_ok=True)

    print("creating new file...")
    get_frontend_db_con(read_only=False)


def write_backend_tables_to_frontend() -> None:
    tables = [
        "int_downloads_melted",
        "int_downloads_melted_daily",
        "int_downloads_melted_weekly",
        "int_downloads_melted_monthly",
        "int_network_stargazers",
        "int_internal_followers",
        "int_network_follower_details",
        "int_status_details",
        "stg_repos",
        "mart_downloads_summary",
        "mart_feed_events",
        "mart_issues",
        "mart_issues_summary",
        "mart_stargazers_pivoted",
        "mart_repo_summary",
        "mart_status_details",
    ]

    backend_con = get_backend_db_con()
    frontend_con = get_frontend_db_con(read_only=False)
    for table in tables:
        offset = 0
        batch_size = 200_000
        first_batch = True
        while True:
            print(f"writing {table} records {offset:,} to {(offset + batch_size):,}...")

            df = backend_con.sql(
                f"SELECT * FROM {table} LIMIT {batch_size} OFFSET {offset}"
            ).to_df()

            if df.empty:
                break

            if first_batch:
                duckdb.sql(
                    f"CREATE TABLE {table} AS SELECT * FROM df", connection=frontend_con
                )
                first_batch = False
            else:
                duckdb.sql(
                    f"INSERT INTO {table} SELECT * FROM df", connection=frontend_con
                )
            offset += batch_size


def write_backend_views_to_frontend() -> None:
    backend_con = get_backend_db_con()
    frontend_con = get_frontend_db_con(read_only=False)

    views_to_copy = [
        "int_status_summary",
        "int_status_summary_pivoted",
        "mart_status_summary",
    ]

    views_to_copy_sql = "'" + "', '".join(views_to_copy) + "'"
    views = backend_con.sql(f"""
        select table_name, view_definition
        from information_schema.views
        where table_schema = 'main'
        and table_name in ({views_to_copy_sql})
    """).fetchall()

    for view_name, view_definition in views:
        view_definition_clean = view_definition.replace("backend", "frontend")
        print(f"creating view {view_name}...")
        frontend_con.sql(f"{view_definition_clean}")


@timeit
def copy_backend_to_frontend() -> None:
    create_new_frontend_db()
    write_backend_tables_to_frontend()
    write_backend_views_to_frontend()


if __name__ == "__main__":
    copy_backend_to_frontend()



from pathlib import Path

from dagster_dbt import DbtProject

ampere_project = DbtProject(
    project_dir=Path(__file__).joinpath("..", "..", "..").resolve(),
    packaged_project_dir=Path(__file__).joinpath("..", "..", "dbt-project").resolve(),
)
ampere_project.prepare_if_dev()



import requests
import typer
from rich.console import Console

from ampere.cli.common import get_api_url
from ampere.models import ReposPublic

console = Console()
repos_app = typer.Typer()


def get_repo_list_response(ctx: typer.Context) -> ReposPublic:
    env = ctx.obj["env"]
    base_url = get_api_url(env)
    url = f"{base_url}/repos/list"

    print(f"Making request to {url}")
    response = requests.get(url)
    assert response.status_code == 200, response.json()
    model = ReposPublic.model_validate(response.json())
    return model


@repos_app.command("list")
def list_repos(ctx: typer.Context) -> None:
    model = get_repo_list_response(ctx)
    console.print_json(model.model_dump_json())



import datetime
import json
import os
import time
from dataclasses import dataclass
from enum import StrEnum, auto
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar

import duckdb
import pandas as pd
import polars as pl
from deltalake import DeltaTable, write_deltalake
from duckdb import DuckDBPyConnection
from sqlmodel import SQLModel
from sqlmodel.main import SQLModelMetaclass

SQLModelType = TypeVar("SQLModelType", bound=SQLModel)


@dataclass
class RefreshConfig:
    model: SQLModelMetaclass
    get_func: Callable


class DeltaTableWriteMode(StrEnum):
    MERGE = auto()
    APPEND = auto()
    OVERWRITE = auto()
    OVERWRITE_WITH_SCHEMA = auto()


@dataclass
class DeltaWriteConfig:
    table_dir: str
    table_name: str
    pks: list[str]
    mode: DeltaTableWriteMode


def format_list_sql_query(input_list: list[str]) -> str:
    return "'" + "', '".join(input_list) + "'"


def create_header(header_length: int, title: str, center: bool, spacer: str):
    if center:
        spacer_len = (header_length - len(title)) // 2
        output = f"{spacer * spacer_len}{title}{spacer * spacer_len}"
    else:
        output = f"{title}{spacer * (header_length - len(title))}"

    if len(output) < header_length:
        output += spacer * (header_length - len(output))
    if len(output) > header_length:
        output = spacer * header_length + "\n" + output

    return output


def get_secret(secret_name: str) -> str:
    import dotenv

    dotenv.load_dotenv()
    secret = os.environ.get(secret_name)
    if secret is None:
        raise ValueError()
    return secret


def get_current_time() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def write_delta_table(
    records: list[SQLModelType] | pd.DataFrame | pl.DataFrame,
    config: DeltaWriteConfig,
    cleanup: bool = True,
) -> None:
    data_dir = Path(__file__).parents[1] / "data" / config.table_dir
    table_path = data_dir / config.table_name
    if isinstance(records, pd.DataFrame):
        df = records
    elif isinstance(records, pl.DataFrame):
        df = records.to_pandas(use_pyarrow_extension_array=True)
    else:
        df = pd.DataFrame.from_records([i.model_dump() for i in records])

    delta_log_dir = table_path / "_delta_log"
    print(f"writing {len(records)} to {table_path}...")
    if not delta_log_dir.exists():
        table_path.mkdir(exist_ok=True, parents=True)
        write_deltalake(table_path, df, mode="error")
        return

    if config.mode == DeltaTableWriteMode.APPEND:
        write_deltalake(table_path, df, mode="append")
        print("append complete")
        return

    elif config.mode == DeltaTableWriteMode.OVERWRITE:
        write_deltalake(table_path, df, mode="overwrite")
        print("overwrite complete")
        return

    elif config.mode == DeltaTableWriteMode.OVERWRITE_WITH_SCHEMA:
        write_deltalake(table_path, df, mode="overwrite", schema_mode="overwrite")
        print("overwrite (including table schema) complete")
        return

    delta_table = DeltaTable(table_path)
    predicate_str = " and ".join([f"s.{i} = t.{i}" for i in config.pks])
    merge_results = (
        delta_table.merge(
            df,
            predicate=predicate_str,
            source_alias="s",
            target_alias="t",
        )
        .when_matched_update_all()
        .when_not_matched_insert_all()
        .execute()
    )
    print(merge_results)

    if cleanup:
        cleanup_delta_table(table_path)


def get_model_primary_key(model: SQLModelMetaclass) -> list[str]:
    pks = []
    for k, v in model.model_fields.items():
        if hasattr(v, "primary_key"):
            pks.append(k)

    return pks


def get_model_foreign_key(model: SQLModelMetaclass, fk_name: str) -> Optional[str]:
    for k, v in model.model_fields.items():
        if not hasattr(v, "foreign_key"):
            continue
        if v.foreign_key == fk_name:
            return k


def get_frontend_db_con(read_only: bool = True) -> DuckDBPyConnection:
    db_path = Path(__file__).parents[1] / "data" / "frontend.duckdb"
    return duckdb.connect(
        str(db_path),
        read_only=read_only,
        config={"TimeZone": "UTC"},
    )


def get_backend_db_con(read_only: bool = True) -> DuckDBPyConnection:
    db_path = Path(__file__).parents[1] / "data" / "backend.duckdb"
    return duckdb.connect(
        str(db_path),
        read_only=read_only,
        config={"TimeZone": "UTC"},
    )


def timeit(func):
    # https://dev.to/kcdchennai/python-decorator-to-measure-execution-time-54hk
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(
            f"{get_current_time()} -- Function {func.__name__} Took {total_time * 1000:.2f} ms"
        )
        return result

    return timeit_wrapper


def vacuum_delta_table(table_path: Path, retention_hours: int = 14 * 24) -> None:
    print("vacuuming...")
    delta_table = DeltaTable(str(table_path))
    results = delta_table.vacuum(retention_hours)
    print("vacuumed", len(results), "files")


def optimize_delta_table(table_path: Path) -> None:
    print("optimizing...")
    delta_table = DeltaTable(str(table_path))
    results = delta_table.optimize.compact()
    print("files added: ", results["numFilesAdded"])
    print("files removed: ", results["numFilesRemoved"])

    if results["numFilesAdded"] > 0:
        size_added = json.loads(results["filesAdded"])["totalSize"]
        size_added_mb = size_added / 1000 / 1000
        print(f"file size (mb) {size_added_mb:.02f}")


def get_delta_tables(base_path: Path) -> list[Path]:
    return [x for x in base_path.iterdir() if x.is_dir()]


def cleanup_delta_table(table_path: Path, retention_hours: int = 14 * 24) -> None:
    optimize_delta_table(table_path)
    vacuum_delta_table(table_path, retention_hours)


def cleanup_delta_tables():
    delta_base_path = Path(__file__).parents[1] / "data" / "bronze"
    all_delta_tables = get_delta_tables(delta_base_path)
    print(create_header(80, "AMPERE MAINTENANCE", True, "="))
    for delta_table_path in all_delta_tables:
        print(create_header(80, delta_table_path.stem, False, "-"))

        optimize_delta_table(delta_table_path)
        vacuum_delta_table(delta_table_path)


def divide_chunks(list_to_chunk: list[Any], n: int):
    # https://stackoverflow.com/a/48135727
    for i in range(0, len(list_to_chunk), n):
        yield list_to_chunk[i : i + n]



from fastapi import FastAPI
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded

from ampere.api.limiter import limiter
from ampere.api.routes import downloads, feed, repos

app = FastAPI()
app.state.limiter = limiter
app.add_exception_handler(
    RateLimitExceeded,
    _rate_limit_exceeded_handler,
)

app.include_router(downloads.router)
app.include_router(feed.router)
app.include_router(repos.router)


@app.get("/")
def root():
    return {"message": "Hello Ampere!"}



from dagster import DefaultScheduleStatus, ScheduleDefinition

from .jobs import bigquery_daily_job, github_metrics_daily_4_job

github_metrics_daily_4 = ScheduleDefinition(
    name="github_metrics_daily_4",
    job=github_metrics_daily_4_job,
    cron_schedule="0 0,6,12,18 * * *",  # every 6 hours, starting at midnight
    default_status=DefaultScheduleStatus.RUNNING,
)

bigquery_daily = ScheduleDefinition(
    name="bigquery_daily",
    job=bigquery_daily_job,
    cron_schedule="0 10 * * *",  # daily 10am utc
    default_status=DefaultScheduleStatus.RUNNING,
)
schedules = [github_metrics_daily_4, bigquery_daily]



from typing import Annotated

import typer

from ampere.cli.commands.downloads import downloads_app
from ampere.cli.commands.feed import feed_app
from ampere.cli.commands.repos import repos_app
from ampere.cli.common import CLIEnvironment

app = typer.Typer(pretty_exceptions_enable=False)
app.add_typer(downloads_app, name="downloads")
app.add_typer(feed_app, name="feed")
app.add_typer(repos_app, name="repos")


@app.callback()
def set_env(
    ctx: typer.Context,
    env: Annotated[CLIEnvironment, typer.Option("--env", "-e")] = CLIEnvironment.prod,
) -> None:
    ctx.obj = {"env": env}


@app.command()
def welcome():
    typer.echo("Welcome to Ampere CLI")
    typer.echo("Use --help to see available commands")


if __name__ == "__main__":
    app()



import datetime
import time
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from typing import Annotated

import requests
import typer
from pydantic import BaseModel
from rich import box
from rich.console import Console
from rich.table import Table

from ampere.cli.common import CLIEnvironment, get_api_url, get_flag_emoji, get_pct_change
from ampere.cli.models import CLIOutputFormat, repo_callback_with_downloads
from ampere.models import (
    DownloadsGranularity,
    DownloadsPublic,
    DownloadsPublicGroup,
    DownloadsSummaryGranularity,
    GetDownloadsPublicConfig,
    create_repo_enum,
)

console = Console()

downloads_app = typer.Typer()


class DownloadsSummary(BaseModel):
    granularity: DownloadsSummaryGranularity
    group: DownloadsPublicGroup
    repo: str
    group_value: str
    min_date: datetime.datetime
    max_date: datetime.datetime
    last_period: int
    this_period: int
    pct_change: float
    pct_total: float


class DownloadsSummaryOutput(BaseModel):
    records: dict[str, list[DownloadsSummary]]
    group: DownloadsPublicGroup
    min_date: datetime.datetime
    max_date: datetime.datetime
    grand_total_this_period: int
    grand_total_last_period: int
    grand_total_pct_change: float


def format_downloads_summary_output(
    config: DownloadsSummaryOutput,
    granularity: DownloadsSummaryGranularity,
    show_grand_total: bool,
) -> Table:
    period = granularity.name.removesuffix("ly")

    title = f"{config.group} downloads summary"
    subtitle = f"{config.min_date.strftime('%Y-%m-%d')} to {config.max_date.strftime('%Y-%m-%d')}"
    title_full = f"{title}\n{subtitle}"

    table = Table(
        title=title_full, title_justify="left", title_style="bold", box=box.ROUNDED
    )
    table.add_column("repo", justify="left")
    table.add_column("group", justify="left")
    table.add_column(f"last {period}", justify="right")
    table.add_column(f"this {period}", justify="right")
    table.add_column("% change", justify="right")

    if config.group != DownloadsPublicGroup.overall:
        table.add_column("% total", justify="right")

    for repo_records in config.records.values():
        if config.group != DownloadsPublicGroup.overall:
            table.add_section()
        for record in repo_records:
            if record.group == "country_code" and len(record.group_value) == 2:
                group_emoji = get_flag_emoji(record.group_value)
                group_value = f"{group_emoji} {record.group_value}"
            else:
                group_value = record.group_value

            row_contents = [
                record.repo,
                group_value,
                f"{record.last_period:,}",
                f"{record.this_period:,}",
                f"{record.pct_change:.2f}%",
            ]
            style = "bold" if record.group_value == "subtotal" else None

            if (
                config.group != DownloadsPublicGroup.overall
                and record.group_value != "subtotal"
            ):
                row_contents.append(f"{record.pct_total:.2f}%")

            table.add_row(*row_contents, style=style)

    if show_grand_total:
        table.add_section()
        table.add_row(
            "total",
            "",
            f"{config.grand_total_last_period:,}",
            f"{config.grand_total_this_period:,}",
            f"{config.grand_total_pct_change:.2f}%",
            style="bold",
        )
    return table


def format_downloads_list_output(response: DownloadsPublic) -> Table:
    repo = response.data[0].repo
    group = response.data[0].group_name

    title = f"{repo} {group} downloads"
    table = Table(
        title=title,
        title_justify="left",
        title_style="bold",
        box=box.ROUNDED,
    )
    table.add_column("timestamp", justify="center")
    table.add_column("group value", justify="left")
    table.add_column("downloads", justify="right")
    if group != DownloadsPublicGroup.overall:
        table.add_column("%", justify="right")

    totals_by_timestamp = {}
    for item in response.data:
        if item.download_timestamp not in totals_by_timestamp:
            totals_by_timestamp[item.download_timestamp] = 0
        totals_by_timestamp[item.download_timestamp] += item.download_count

    prev_timestamp = response.data[0].download_timestamp
    for item in response.data:
        if (
            item.download_timestamp != prev_timestamp
            and group != DownloadsPublicGroup.overall
        ):
            table.add_section()
        prev_timestamp = item.download_timestamp

        pct = item.download_count / totals_by_timestamp[item.download_timestamp] * 100

        if group == "country_code":
            group_emoji = get_flag_emoji(item.group_value)
            group_value = f"{group_emoji} {item.group_value}"
        else:
            group_value = item.group_value

        row_contents = [
            item.download_timestamp.strftime("%Y-%m-%d %H:%M:%S"),
            group_value,
            f"{item.download_count:,}",
        ]
        if group != DownloadsPublicGroup.overall:
            row_contents.append(str(round(pct, 2)))

        table.add_row(*row_contents)

    return table


def get_downloads_response(
    config: GetDownloadsPublicConfig, ctx: typer.Context
) -> DownloadsPublic:
    env = ctx.obj["env"]
    if env == CLIEnvironment.dev:
        time.sleep(0.4)

    base_url = get_api_url(env)
    url = f"{base_url}/downloads/{config.granularity}"
    response = requests.get(
        url,
        params={
            "repo": config.repo,
            "group": config.group,
            "n_days": config.n_days,
            "limit": config.limit,
            "descending": config.descending,
        },
    )
    assert response.status_code == 200, print(response.json())
    return DownloadsPublic.model_validate(response.json())


@downloads_app.command("list")
def list_downloads(
    ctx: typer.Context,
    granularity: Annotated[
        DownloadsGranularity,
        typer.Option(
            "--granularity",
            "-g",
            prompt=True,
        ),
    ],
    repo: Annotated[
        str,
        typer.Option("--repo", "-r", prompt=True, callback=repo_callback_with_downloads),
    ],
    group: Annotated[
        DownloadsPublicGroup, typer.Option("--group", "-gr")
    ] = DownloadsPublicGroup.overall,
    n_days: Annotated[int, typer.Option("--n-days", "-n")] = 180,
    limit: Annotated[int, typer.Option("--limit", "-l")] = 10_000,
    descending: Annotated[bool, typer.Option("--desc/--asc", "-d/-a")] = True,
    output: Annotated[
        CLIOutputFormat, typer.Option("--output", "-o")
    ] = CLIOutputFormat.table,
):
    response = get_downloads_response(
        GetDownloadsPublicConfig(
            granularity=granularity,
            repo=repo,
            group=group,
            n_days=n_days,
            limit=limit,
            descending=descending,
        ),
        ctx=ctx,
    )
    if output == CLIOutputFormat.json:
        console.print_json(response.model_dump_json())
        return

    table = format_downloads_list_output(response)
    console.print(table)


def create_downloads_summary(
    records: list[DownloadsPublic],
    group: DownloadsPublicGroup,
    granularity: DownloadsSummaryGranularity,
    descending: bool,
    min_pct_of_total: float,
    show_subtotal: bool,
    ctx: typer.Context,
) -> DownloadsSummaryOutput:
    env = ctx.obj["env"]
    others: dict[create_repo_enum(env, True), DownloadsSummary] = {}  # type: ignore
    repo_summaries: dict[create_repo_enum(env, True), list[DownloadsSummary]] = {}  # type: ignore

    grand_total_this_period = 0
    grand_total_last_period = 0
    min_dates = []
    max_dates = []

    for repo_data in records:
        summary = []
        group_val_lookup = {}
        repo = repo_data.data[0].repo
        for item in repo_data.data:
            if item.group_value not in group_val_lookup:
                group_val_lookup[item.group_value] = {
                    "min_date": item.download_timestamp,
                    "max_date": item.download_timestamp,
                    "this_period": item.download_count,
                    "last_period": 0,
                }
            else:
                group_val_lookup[item.group_value].update(
                    {
                        "min_date": item.download_timestamp,
                        "last_period": item.download_count,
                    }
                )

        total_this_period = 0
        total_last_period = 0
        total_min_date = None
        total_max_date = None
        for val in group_val_lookup.values():
            total_this_period += val["this_period"]
            total_last_period += val["last_period"]

            if total_min_date is None or val["min_date"] < total_min_date:
                total_min_date = val["min_date"]
            if total_max_date is None or val["max_date"] > total_max_date:
                total_max_date = val["max_date"]

        assert all([total_min_date, total_max_date])
        grand_total_last_period += total_last_period
        grand_total_this_period += total_this_period
        min_dates.append(total_min_date)
        max_dates.append(total_max_date)

        subtotal_record = DownloadsSummary(
            granularity=granularity,
            group=group,
            group_value="subtotal",
            repo=repo,
            last_period=total_last_period,
            this_period=total_this_period,
            min_date=min([i["min_date"] for i in group_val_lookup.values()]),
            max_date=max([i["max_date"] for i in group_val_lookup.values()]),
            pct_change=get_pct_change(total_last_period, total_this_period),
            pct_total=100,
        )

        for k, v in group_val_lookup.items():
            last_period = v["last_period"]
            this_period = v["this_period"]

            pct_change = get_pct_change(last_period, this_period)
            pct_total = this_period / total_this_period * 100

            record_parsed = DownloadsSummary(
                granularity=granularity,
                group=group,
                group_value=k,
                repo=repo,
                last_period=last_period,
                this_period=this_period,
                min_date=v["min_date"],
                max_date=v["max_date"],
                pct_change=pct_change,
                pct_total=pct_total,
            )
            if pct_total < min_pct_of_total:
                record_parsed.group_value = "other"
                if repo not in others:
                    others[repo] = record_parsed
                    continue
                others[repo].this_period += this_period
                others[repo].last_period += last_period

            else:
                summary.append(record_parsed)

        if descending:
            summary.insert(0, subtotal_record)
        else:
            summary.append(subtotal_record)

        repo_summaries[repo] = summary

    for k, v in others.items():
        pct_change = get_pct_change(v.last_period, v.this_period)
        v.pct_change = pct_change

    summary_totals = []
    for k, v in repo_summaries.items():
        subtotal_index = 0 if descending else -1
        summary_totals.append((k, v[subtotal_index].this_period))

        if not show_subtotal:
            _ = v.pop(subtotal_index)

        # final group level sort to account for new "other" category
        if k in others:
            v.append(others[k])
            v.sort(key=lambda x: x.this_period, reverse=descending)

    # top level repo level sort by total downloads
    summary_totals.sort(key=lambda x: x[1], reverse=descending)

    sorted_repo_summaries = {}
    for i in summary_totals:
        sorted_repo_summaries[i[0]] = repo_summaries[i[0]]

    return DownloadsSummaryOutput(
        records=sorted_repo_summaries,
        group=group,
        min_date=min(min_dates),
        max_date=max(max_dates),
        grand_total_last_period=grand_total_last_period,
        grand_total_this_period=grand_total_this_period,
        grand_total_pct_change=get_pct_change(
            grand_total_last_period, grand_total_this_period
        ),
    )


@downloads_app.command("summary")
def summarize_downloads(
    ctx: typer.Context,
    granularity: Annotated[
        DownloadsSummaryGranularity,
        typer.Option(
            "--granularity",
            "-g",
            prompt=True,
        ),
    ],
    repo: Annotated[
        str | None,
        typer.Option("--repo", "-r", callback=repo_callback_with_downloads),
    ] = None,
    group: Annotated[
        DownloadsPublicGroup, typer.Option("--group", "-gr")
    ] = DownloadsPublicGroup.overall,
    descending: Annotated[bool, typer.Option("--desc/--asc", "-d/-a")] = True,
    min_group_pct_of_total: Annotated[
        float,
        typer.Option(
            "--min-pct",
            "-m",
            help="""
            The minimum download count % of total in the most recent period for the specified repo and group.
            Group values with a % total lower than this threshold will be collapsed into an 'other' category.
            """,
        ),
    ] = 0.1,
    show_subtotal: Annotated[bool, typer.Option("--show-subtotal", "-st")] = False,
    show_grand_total: Annotated[bool, typer.Option("--show-grand-total", "-gt")] = False,
    output: Annotated[
        CLIOutputFormat, typer.Option("--output", "-o")
    ] = CLIOutputFormat.table,
):
    filters = {
        granularity.monthly: {"limit": 10000, "n_days": 7 * 4 * 3},
        granularity.weekly: {"limit": 10000, "n_days": 7 * 3},
    }

    env = ctx.obj["env"]
    if repo is None:
        repos = [i for i in create_repo_enum(env, True)]
    else:
        repos = [repo]

    configs = []
    for repo in repos:
        configs.append(
            GetDownloadsPublicConfig(
                granularity=granularity,
                repo=repo,
                group=group,
                n_days=filters[granularity]["n_days"],
                limit=filters[granularity]["limit"],
                descending=True,  # get the most recent data regardless of the sort order
            )
        )

    with ThreadPoolExecutor(max_workers=len(repos)) as executor:
        all_records = list(executor.map(get_downloads_response, configs, repeat(ctx)))

    summary: DownloadsSummaryOutput = create_downloads_summary(
        records=all_records,
        group=group,
        granularity=granularity,
        descending=descending,
        min_pct_of_total=min_group_pct_of_total,
        show_subtotal=show_subtotal,
        ctx=ctx,
    )
    if output == CLIOutputFormat.json:
        for _, record_list in summary.records.items():
            for record in record_list:
                console.print_json(record.model_dump_json())
        return

    table = format_downloads_summary_output(
        config=summary, granularity=granularity, show_grand_total=show_grand_total
    )
    console.print(table)



from fastapi import APIRouter, Request

from ampere.api.limiter import limiter
from ampere.cli.common import CLIEnvironment
from ampere.models import ReposPublic, get_repo_names

router = APIRouter(prefix="/repos", tags=["repo"])


@limiter.limit("60/minute")
@router.get("/list", response_model=ReposPublic)
def read_repos(request: Request) -> ReposPublic:
    repos = get_repo_names(CLIEnvironment.dev)
    print(repos)
    return ReposPublic(repos=repos, count=len(repos))



import pickle
import random
from pathlib import Path
from typing import Any

import networkx as nx
import pandas as pd
from plotly.graph_objs import Figure

from ampere.cli.common import CLIEnvironment
from ampere.common import get_backend_db_con, timeit
from ampere.models import Followers, StargazerNetworkRecord, get_repos_with_downloads
from ampere.styling import ScreenWidth
from ampere.viz import (
    get_downloads_data,
    get_summary_data,
    viz_downloads,
    viz_follower_network,
    viz_star_network,
    viz_summary,
)

SCREEN_WIDTHS = [
    ScreenWidth.xs,
    ScreenWidth.sm,
    ScreenWidth.md,
    ScreenWidth.lg,
    ScreenWidth.xl,
]

MODES = ["light", "dark"]


def dump_obj_to_pickle(pkl_name: str, obj: Any):
    out_dir = Path(__file__).parents[1] / "data" / "viz"
    out_dir.mkdir(exist_ok=True, parents=True)

    out_path = out_dir / f"{pkl_name}.pkl"
    with out_path.open("wb") as f:
        pickle.dump(obj, f)


def dump_fig_to_json(f_name: str, fig: Figure):
    out_dir = Path(__file__).parents[1] / "data" / "viz"
    out_dir.mkdir(exist_ok=True, parents=True)

    out_path = out_dir / f"{f_name}.json"
    serialized_fig = fig.to_json()
    if not isinstance(serialized_fig, str):
        raise TypeError()

    with out_path.open("w") as f:
        f.write(serialized_fig)


def create_follower_network() -> None:
    con = get_backend_db_con()
    followers = (
        con.sql("select user_id, follower_id from int_internal_followers")
        .to_df()
        .to_dict(orient="records")
    )
    followers = list(Followers(**record) for record in followers)  # type: ignore

    random.seed(42)
    added_nodes = []
    graph = nx.Graph()
    for record in followers:
        if record.user_id not in added_nodes:
            graph.add_node(record.user_id)
            added_nodes.append(record.user_id)

        if record.follower_id not in added_nodes:
            graph.add_node(record.follower_id)
            added_nodes.append(record.follower_id)

        graph.add_edge(record.user_id, record.follower_id, weight=0.03)
    pos = nx.spring_layout(graph)
    nx.set_node_attributes(graph, pos, "pos")

    dump_obj_to_pickle("follower_network", graph)


@timeit
def cache_summary_plots() -> None:
    df = pd.DataFrame(get_summary_data())
    metrics = ["stars", "issues", "commits"]

    for metric in metrics:
        for mode in MODES:
            dark_mode = mode == "dark"
            for width in SCREEN_WIDTHS:
                fig = viz_summary(
                    df,
                    metric_type=metric,
                    date_range=None,
                    screen_width=width,
                    dark_mode=dark_mode,
                )

                f_name = f"summary_{metric}_{mode}_{width.value}"
                dump_obj_to_pickle(f_name, fig)


@timeit
def cache_downloads_plots() -> None:
    repos = get_repos_with_downloads(CLIEnvironment.dev)
    groups = ["overall", "package_version", "python_version"]
    for repo in repos:
        df = get_downloads_data(repo)
        dump_obj_to_pickle(f"downloads_df_{repo}", df)
        for group in groups:
            for mode in MODES:
                for width in SCREEN_WIDTHS:
                    dark_mode = mode == "dark"
                    pkl_name = f"downloads_{repo}_{group}_{mode}_{width}"
                    print(f"caching {pkl_name}...")
                    fig = viz_downloads(
                        df,
                        group_name=group,
                        date_range=None,
                        dark_mode=dark_mode,
                        screen_width=width,
                    )
                    dump_obj_to_pickle(pkl_name, fig)


def cache_stargazer_network():
    for mode in MODES:
        for width in SCREEN_WIDTHS:
            dark_mode = mode == "dark"
            fig = viz_star_network(dark_mode, width)
            f_name = f"stargazer_network_{mode}_{width}"
            dump_obj_to_pickle(f_name, fig)


def cache_follower_network():
    for mode in MODES:
        for width in SCREEN_WIDTHS:
            dark_mode = mode == "dark"
            fig = viz_follower_network(dark_mode, width)
            f_name = f"follower_network_{mode}_{width}"
            dump_obj_to_pickle(f_name, fig)


@timeit
def create_stargazer_network() -> None:
    con = get_backend_db_con()
    stargazers_df = con.sql(
        """
        select
            user_name,
            followers_count,
            starred_at,
            retrieved_at,
            repo_name
        from int_network_stargazers
    """
    ).to_df()

    stargazers = list(StargazerNetworkRecord(*record) for record in stargazers_df.values)
    repos_with_stargazers = list(set(i.repo_name for i in stargazers))

    print("creating star network...")
    random.seed(42)
    added_repos = []
    current_user = stargazers[0].user_name

    graph = nx.Graph()
    for repo in repos_with_stargazers:
        graph.add_node(repo, node_type="repo", repo=repo)

    for record in stargazers:
        if record.user_name != current_user:
            added_repos = []
        node_name = f"{record.user_name}_{record.repo_name}"
        graph.add_node(
            node_name,
            followers_count=record.followers_count,
            node_type="user_repo",
            repo=record.repo_name,
        )
        graph.add_edge(node_name, record.repo_name, weight=50, edge_type="user_repo")
        if len(added_repos) > 0:
            for added_repo in added_repos:
                graph.add_edge(
                    node_name,
                    f"{record.user_name}_{added_repo}",
                    edge_type="user_user",
                    weight=0.1,
                )
        added_repos.append(record.repo_name)
        current_user = record.user_name

    pos = nx.spring_layout(graph)
    nx.set_node_attributes(graph, pos, "pos")

    dump_obj_to_pickle("star_network", graph)


def refresh_all_caches():
    create_follower_network()
    cache_follower_network()

    create_stargazer_network()
    cache_stargazer_network()

    cache_summary_plots()
    cache_downloads_plots()


if __name__ == "__main__":
    refresh_all_caches()



from dataclasses import asdict, dataclass
from enum import StrEnum, auto
from typing import Any

import pandas as pd

from ampere.common import timeit


class AmperePalette(StrEnum):
    PAGE_BACKGROUND_COLOR_LIGHT = "rgb(242, 240, 227)"
    BRAND_TEXT_COLOR_LIGHT = "rgb(33, 33, 33)"
    PAGE_BACKGROUND_COLOR_DARK = "rgb(33, 33, 33)"
    BRAND_TEXT_COLOR_DARK = "rgb(242, 240, 227)"


class ScreenWidth(StrEnum):
    xs = auto()
    sm = auto()
    md = auto()
    lg = auto()
    xl = auto()


@dataclass
class DTStyle:
    sort_action: str
    sort_mode: str
    column_selectable: str
    row_selectable: bool
    row_deletable: bool
    fixed_rows: dict[str, bool]
    filter_options: dict[str, str]
    page_size: int
    style_header: dict[str, str]
    filter_action: str
    style_filter: dict[str, str]
    style_cell: dict[str, Any]
    style_cell_conditional: list[dict[str, Any]]
    style_data_conditional: list[dict[str, Any]]
    style_data: dict[str, str]
    style_table: dict[str, Any]
    css: list[dict[str, Any]]


@dataclass
class ColumnInfo:
    name: str
    ascending: bool
    palette: str


def adjust_rgb_for_dark_mode(rgb: str) -> str:
    rgb_vals = rgb.split("rgb(")[1].split(")")[0].split(",")
    adjusted_rgb = [str(255 - int(i)) for i in rgb_vals]
    return f'rgb({",".join(adjusted_rgb)})'


def generate_heatmap_palette(dark_mode: bool, palette_name: str) -> list[str]:
    if palette_name not in ["oranges", "greens"]:
        raise ValueError("Invalid palette name")

    palettes = {
        "oranges": {
            "light": ["#FFC6BE", "#FF9182", "#FF4B33"],
            "dark": ["#660C00", "#881000", "#CC1800"],
        },
        "greens": {
            "light": [
                "#C8F6B2",
                "#8DEC60",
                "#57D61A",
            ],
            "dark": ["#112A05", "#22530A", "#337D0F"],
        },
    }

    return palettes[palette_name]["dark" if dark_mode else "light"]


@timeit
def style_dt_background_colors_by_rank(
    df: pd.DataFrame, n_bins: int, cols: list[ColumnInfo], dark_mode: bool
) -> list[dict]:
    # https://dash.plotly.com/datatable/conditional-formatting
    styles = []
    rows_to_update = 3
    _, color = get_ampere_colors(dark_mode, contrast=False)

    for col in cols:
        colors = generate_heatmap_palette(dark_mode, col.palette)
        ranks = (
            (df[col.name].rank(ascending=col.ascending, method="max").astype("int") - 1)
            .squeeze()
            .tolist()  # type: ignore
        )
        max_ranks = sorted(set(ranks))[-rows_to_update:]
        for row in range(n_bins):
            row_val = df[col.name].iloc[row]
            rank_val = ranks[row]
            if rank_val not in max_ranks:
                continue

            rank_val_idx = max_ranks.index(rank_val)
            styles.append(
                {
                    "if": {
                        "filter_query": f"{{{col.name}}} = {row_val}",
                        "column_id": col.name,
                    },
                    "backgroundColor": colors[rank_val_idx],
                    "color": color,
                    "borderLeft": f"2px solid {color}",
                }
            )

    return styles


def get_ampere_dt_style(dark_mode: bool = False) -> dict:
    background, color = get_ampere_colors(dark_mode, contrast=False)
    dt_style = asdict(
        DTStyle(
            sort_action="native",
            sort_mode="multi",
            column_selectable="single",
            row_selectable=False,
            row_deletable=False,
            fixed_rows={"headers": True},
            filter_action="native",
            filter_options={"case": "insensitive", "placeholder_text": ""},
            page_size=100,
            style_header={},
            style_filter={},
            style_cell={
                "textAlign": "center",
                "minWidth": 100,
                "maxWidth": 170,
                "whiteSpace": "normal",
                "wordBreak": "break-word",
            },
            style_cell_conditional=[],
            style_data_conditional=[],
            style_data={
                "color": color,
                "backgroundColor": background,
            },
            css=[],
            style_table={
                "height": "85vh",
                "maxHeight": "85vh",
                "overflowY": "scroll",
                "overflowX": "scroll",
            },
        )
    )
    return dt_style


def get_ampere_colors(dark_mode: bool, contrast: bool) -> tuple[str, str]:
    colors = {
        "dark": {
            "background": AmperePalette.PAGE_BACKGROUND_COLOR_DARK,
            "text": AmperePalette.BRAND_TEXT_COLOR_DARK,
        },
        "light": {
            "background": AmperePalette.PAGE_BACKGROUND_COLOR_LIGHT,
            "text": AmperePalette.BRAND_TEXT_COLOR_LIGHT,
        },
    }

    # invert the selector if contrast is True
    selector = not dark_mode if contrast else dark_mode
    selector_str = "dark" if selector else "light"

    return colors[selector_str]["background"], colors[selector_str]["text"]


def get_table_title_style(dark_mode: bool) -> dict:
    background, color = get_ampere_colors(dark_mode, contrast=True)

    return {
        "color": color,
        "backgroundColor": background,
        "paddingBottom": "0",
        "paddingLeft": "10px",
        "paddingRight": "10px",
        "fontSize": "1.5rem",
        "fontWeight": "bold",
        "marginBottom": "0",
        "border": "none",
    }



import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dash_table, html

from ampere.common import get_frontend_db_con, timeit
from ampere.styling import (
    ScreenWidth,
    get_ampere_colors,
    get_ampere_dt_style,
)


@timeit
def create_repo_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
        with downloads_total as (
            select
                repo,
                sum(download_count) as total_downloads
            from mart_downloads_summary
            where group_name = 'overall'
            group by all
        ),
        downloads_past_week as (
            select
                repo,
                download_count as downloads_past_week
            from mart_downloads_summary
            where group_name = 'overall'
            and download_date = (select max(download_date) from mart_downloads_summary) 
       )
        select
            concat('[', a.repo_name, ']', '(https://www.github.com/mrpowers-io/', a.repo_name, ')') as repo_name,
            a.forks_count                                                                           as forks,
            a.stargazers_count                                                                      as stargazers,
            a.open_issues_count                                                                     as "open issues",
            round(date_part('day', current_date - a.created_at) / 365, 1)                           as "age (years)",
            strftime(a.created_at, '%Y-%m-%d')                                                      as created,
            strftime(a.updated_at, '%Y-%m-%d')                                                      as updated,
            b.total_downloads                                                                       as "downloads (total)",
            c.downloads_past_week                                                                   as "downloads (past week)"
        from stg_repos a
        left join downloads_total as b on a.repo_name = b.repo
        left join downloads_past_week as c on a.repo_name = c.repo
        order by a.stargazers_count desc
        """,
        ).to_df()

    return df


@callback(
    [
        Output("about-text", "children"),
        Output("about-text", "style"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def get_styled_about_text(dark_mode: bool, breakpoint_name: str):
    _, color = get_ampere_colors(dark_mode, False)

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        font_size = "12px"
        margin_left = "0vw"
    else:
        font_size = "14px"
        margin_left = "20vw"

    return [
        html.Div(
            id="about-text",
            children=[
                html.Div(
                    "made with ❤️ by ",
                    style={"display": "inline", "color": color, "fontSize": font_size},
                ),
                html.A(
                    "Jeff Brennan",
                    href="https://github.com/jeffbrennan",
                    target="_blank",
                    style={"fontSize": font_size},
                ),
                html.Div(
                    " and ",
                    style={"display": "inline", "color": color, "fontSize": font_size},
                ),
                html.A(
                    "mrpowers-io",
                    href="https://github.com/mrpowers-io",
                    target="_blank",
                    style={"fontSize": font_size},
                ),
                html.Div(
                    " contributors",
                    style={"display": "inline", "color": color, "fontSize": font_size},
                ),
            ],
            style={"marginLeft": margin_left},
        ),
    ], {}


@callback(
    [
        Output("about-table", "children"),
        Output("about-table", "style"),
        Output("about-fade", "is_in"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def get_styled_about_table(dark_mode: bool, breakpoint_name: str):
    df = create_repo_table()
    about_style = get_ampere_dt_style(dark_mode)
    about_style["style_table"]["height"] = "auto"

    sm_margins = {"maxWidth": "90vw", "width": "90vw"}
    lg_margins = {"maxWidth": "50vw", "width": "50vw", "marginLeft": "20vw"}

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        about_style["style_cell"]["font_size"] = "12px"
        about_style["style_table"].update(sm_margins)

    else:
        about_style["style_table"].update(lg_margins)

    tbl_cols = []
    for col in df.columns:
        if col == "repo_name":
            tbl_cols.append(
                {"id": "repo_name", "name": "repo", "presentation": "markdown"}
            )
        elif "downloads" in col:
            tbl_cols.append(
                {"name": col, "id": col, "type": "numeric", "format": {"specifier": ",d"}}
            )
        else:
            tbl_cols.append({"id": col, "name": col})

    tbl = dash_table.DataTable(
        df.to_dict("records"),
        columns=tbl_cols,
        **about_style,
    )

    return tbl, {}, True


def layout():
    return [
        dbc.Fade(
            id="about-fade",
            children=[
                html.Div(id="about-table", style={"visibility": "hidden"}),
                html.Br(),
                html.Div(id="about-text", style={"visibility": "hidden"}),
            ],
            style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
            is_in=False,
        )
    ]



import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dash_table, html

from ampere.common import get_frontend_db_con, timeit
from ampere.styling import ScreenWidth, get_ampere_dt_style


def create_status_summary_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
        select
            summary,
            downloads,
            feed,
            issues,
            network_stargazers as "network stargazers",
            network_followers as "network followers"
        from 
        mart_status_summary
        """
        ).to_df()
    return df


def create_status_details_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        df = con.sql(
            """
        select
            concat(
                '[', replace(model, '_', ' '), ']',
                '(https://github.com/jeffbrennan/ampere/tree/main/models/',
                 model_folder, '/', "model", '.sql)'
            ) as "model",
            array_to_string(page, ', ') as pages,
            timestamp_col,
            case when "timestamp" is not null
                then concat(strftime(timestamp, '%Y-%m-%d %H:%M:%S'), ' (UTC)')
            end "timestamp",
            round((extract(epoch from now()) - extract(epoch from "timestamp")) / 3600, 2) as hours_stale,
            records
            from mart_status_details
        """,
        ).to_df()

    return df


@callback(
    Output("status-fade", "is_in"),
    Input("status-table", "id"),
)
def status_table_fadein(_: str) -> bool:
    return True


@callback(
    [
        Output("status-summary-table", "children"),
        Output("status-summary-table", "style"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
@timeit
def get_styled_summary_table(dark_mode: bool, breakpoint_name: str):
    summary_df = create_status_summary_table()
    summary_style = get_ampere_dt_style(dark_mode)
    summary_style["filter_action"] = "none"
    summary_style["sort_action"] = "none"

    lg_margins = {
        "maxWidth": "50vw",
        "width": "50vw",
        "marginLeft": "20vw",
    }

    sm_margins = {
        "maxWidth": "90vw",
        "width": "90vw",
        "marginLeft": "0vw",
    }

    summary_style["style_table"]["height"] = "25%"
    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        summary_style["style_table"].update(sm_margins)
        summary_style["style_cell"]["font_size"] = "12px"

    else:
        summary_style["style_table"].update(lg_margins)

    tbl = dash_table.DataTable(
        data=summary_df.to_dict("records"),
        columns=[{"id": x, "name": x} for x in summary_df.columns],
        **summary_style,
    )

    print("updating summary table")
    return tbl, {}


@callback(
    [
        Output("status-details-table", "children"),
        Output("status-details-table", "style"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
def get_styled_details_table(dark_mode: bool, breakpoint_name: str):
    details_df = create_status_details_table()
    print(details_df.head())
    details_cols = []
    for col in details_df.columns:
        if col in ["model"]:
            details_cols.append({"name": col, "id": col, "presentation": "markdown"})
        elif col == "records":
            details_cols.append(
                {"name": col, "id": col, "type": "numeric", "format": {"specifier": ",d"}}
            )
        else:
            details_cols.append({"name": col, "id": col})

    status_details_style = get_ampere_dt_style(dark_mode)
    lg_margins = {
        "maxWidth": "50vw",
        "width": "50vw",
        "marginLeft": "20vw",
    }

    sm_margins = {
        "maxWidth": "90vw",
        "width": "90vw",
        "marginLeft": "0vw",
    }

    status_details_style["style_table"].update({"height": "85vh", "maxHeight": "85vh"})
    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        status_details_style["style_table"].update(sm_margins)
        status_details_style["style_cell"]["font_size"] = "12px"

    else:
        status_details_style["style_table"].update(lg_margins)

    status_details_style["style_cell"]["textAlign"] = "left"
    tbl = (
        dash_table.DataTable(
            details_df.to_dict("records"),
            columns=details_cols,
            id="status-table",
            **status_details_style,
        ),
    )

    return tbl, {}


def layout():
    return [
        dbc.Fade(
            id="status-fade",
            children=[
                html.Div(id="status-summary-table", style={"visibility": "hidden"}),
                html.Br(),
                html.Div(id="status-details-table", style={"visibility": "hidden"}),
            ],
            style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
            is_in=False,
        ),
    ]



from fastapi import APIRouter, Query, Request
from pydantic import TypeAdapter

from ampere.api.limiter import limiter
from ampere.cli.common import CLIEnvironment
from ampere.common import get_frontend_db_con
from ampere.models import (
    DownloadPublic,
    DownloadsGranularity,
    DownloadsPublic,
    DownloadsPublicGroup,
    GetDownloadsPublicConfig,
    ReposWithDownloads,
    create_repo_enum,
)

router = APIRouter(prefix="/downloads", tags=["downloads"])


def get_downloads_base(
    table_name: str, config: GetDownloadsPublicConfig
) -> DownloadsPublic:
    con = get_frontend_db_con()
    valid_repos = create_repo_enum(CLIEnvironment.dev, True)
    try:
        repo = valid_repos(config.repo)  # type: ignore
    except ValueError:
        raise ValueError(f"Invalid repo: {config.repo}")

    params = [repo.value, config.group]
    sort_order = "desc" if config.descending else "asc"

    query = f"""
    select * 
    from {table_name}
    where repo = ?
    and group_name = ?
    and download_timestamp >= now() - interval {config.n_days} days
    order by download_timestamp {sort_order}, download_count {sort_order}
    limit {config.limit}
    """

    result = con.sql(query, params=params).to_df().to_dict(orient="records")
    ta = TypeAdapter(list[DownloadPublic])
    return DownloadsPublic(data=ta.validate_python(result), count=len(result))


@router.get("/hourly", response_model=DownloadsPublic)
@limiter.limit("60/minute")
def read_downloads_hourly(
    request: Request,
    repo: str,
    group: DownloadsPublicGroup = DownloadsPublicGroup.overall,
    n_days: int = Query(default=7, le=24 * 365 * 5),
    limit: int = Query(default=7 * 24, le=100_000),
    descending: bool = Query(default=True),
) -> DownloadsPublic:
    return get_downloads_base(
        table_name="int_downloads_melted",
        config=GetDownloadsPublicConfig(
            granularity=DownloadsGranularity.hourly,
            repo=repo,
            group=group,
            n_days=n_days,
            limit=limit,
            descending=descending,
        ),
    )


@router.get("/daily", response_model=DownloadsPublic)
@limiter.limit("60/minute")
def read_downloads_daily(
    request: Request,
    repo: str,
    group: DownloadsPublicGroup = DownloadsPublicGroup.overall,
    n_days: int = Query(default=30, le=365 * 5),
    limit: int = Query(default=100, le=100_000),
    descending: bool = Query(default=True),
) -> DownloadsPublic:
    return get_downloads_base(
        table_name="int_downloads_melted_daily",
        config=GetDownloadsPublicConfig(
            granularity=DownloadsGranularity.daily,
            repo=repo,
            group=group,
            n_days=n_days,
            limit=limit,
            descending=descending,
        ),
    )


@router.get("/weekly", response_model=DownloadsPublic)
@limiter.limit("60/minute")
def read_downloads_weekly(
    request: Request,
    repo: str,
    group: DownloadsPublicGroup = DownloadsPublicGroup.overall,
    n_days: int = Query(default=30, le=365 * 5),
    limit: int = Query(default=100, le=10_000),
    descending: bool = Query(default=True),
) -> DownloadsPublic:
    return get_downloads_base(
        table_name="int_downloads_melted_weekly",
        config=GetDownloadsPublicConfig(
            granularity=DownloadsGranularity.weekly,
            repo=repo,
            group=group,
            n_days=n_days,
            limit=limit,
            descending=descending,
        ),
    )


@router.get("/monthly", response_model=DownloadsPublic)
@limiter.limit("60/minute")
def read_downloads_monthly(
    request: Request,
    repo: str,
    group: DownloadsPublicGroup = DownloadsPublicGroup.overall,
    n_days: int = Query(default=60, le=365 * 5),
    limit: int = Query(default=100, le=10_000),
    descending: bool = Query(default=True),
) -> DownloadsPublic:
    return get_downloads_base(
        table_name="int_downloads_melted_monthly",
        config=GetDownloadsPublicConfig(
            granularity=DownloadsGranularity.monthly,
            repo=repo,
            group=group,
            n_days=n_days,
            limit=limit,
            descending=descending,
        ),
    )


@router.get("/repos", response_model=ReposWithDownloads)
@limiter.limit("60/minute")
def read_repos_with_downloads(request: Request) -> ReposWithDownloads:
    con = get_frontend_db_con()
    query = """
        select distinct a.repo 
        from mart_downloads_summary a 
        left join stg_repos b on a.repo = b.repo_name
        order by b.stargazers_count desc, a.repo
    """
    repos = con.sql(query).to_df().squeeze().tolist()
    return ReposWithDownloads(repos=repos, count=len(repos))



import os

import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dash_table, dcc, html

from ampere.common import get_frontend_db_con, timeit
from ampere.styling import ScreenWidth, get_ampere_dt_style
from ampere.viz import read_plotly_fig_pickle, viz_star_network


def create_stargazers_table() -> pd.DataFrame:
    with get_frontend_db_con() as con:
        # select * because columns are dynamically generated
        df = con.sql(
            """
        select *
        from mart_stargazers_pivoted
        order by followers desc
        """,
        ).to_df()
    return df


@callback(
    [
        Output("network-stargazer-graph", "figure"),
        Output("network-stargazer-graph", "style"),
        Output("network-stargazer-table", "style"),
        Output("network-stargazer-graph-fade", "is_in"),
    ],
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
@timeit
def get_stylized_network_graph(dark_mode: bool, breakpoint_name: str):
    mode = "dark" if dark_mode else "light"
    env = os.environ.get("AMPERE_ENV")

    if env == "prod":
        fname = f"stargazer_network_{mode}_{breakpoint_name}"
        fig = read_plotly_fig_pickle(fname)
    else:
        fig = viz_star_network(
            dark_mode=dark_mode,
            screen_width=ScreenWidth(breakpoint_name),
        )

    return (
        fig,
        {
            "height": "95vh",
            "marginLeft": "0vw",
            "marginRight": "0vw",
            "width": "100%",
        },
        {},
        True,
    )


@callback(
    Output("network-stargazer-table", "children"),
    [
        Input("color-mode-switch", "value"),
        Input("breakpoints", "widthBreakpoint"),
    ],
)
@timeit
def get_styled_stargazers_table(dark_mode: bool, breakpoint_name: str):
    base_style = get_ampere_dt_style(dark_mode)
    df = create_stargazers_table()
    if dark_mode:
        text_color = "white"
    else:
        text_color = "black"

    standard_col_colors = [
        {
            "color": text_color,
            "borderLeft": "none",
            "borderRight": f"2px solid {text_color}",
        }
        for _ in df.columns
    ]
    base_style["style_data_conditional"] += standard_col_colors

    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        base_style["style_cell"]["font_size"] = "12px"

    tbl = (
        dash_table.DataTable(
            df.to_dict("records"),
            columns=[
                (
                    {"id": x, "name": "", "presentation": "markdown"}
                    if x == "user_name"
                    else {"id": x, "name": x}
                )
                for x in df.columns
            ],
            **base_style,
        ),
    )
    return tbl


def layout():
    return [
        dbc.Fade(
            id="network-stargazer-graph-fade",
            children=[
                dcc.Graph(
                    id="network-stargazer-graph",
                    style={"visibility": "hidden"},
                    responsive=True,
                    config={"displayModeBar": False},
                ),
                html.Div(id="network-stargazer-table", style={"visibility": "hidden"}),
            ],
            style={
                "transition": "opacity 200ms ease-in",
                "minHeight": "100vh",
            },
            is_in=False,
        ),
    ]
